from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Literal, Optional, Union

from pydantic import BaseModel, Field


class TogetherJSONModel(BaseModel):
    """Base model with JSON serialization support."""

    def model_dump(self, **kwargs: Any) -> Dict[str, Any]:
        exclude_none = kwargs.pop("exclude_none", True)
        data = super().model_dump(exclude_none=exclude_none, **kwargs)

        # Convert datetime objects to ISO format strings
        for key, value in data.items():
            if isinstance(value, datetime):
                data[key] = value.isoformat()

        return data


class Autoscaling(TogetherJSONModel):
    """Configuration for automatic scaling of replicas based on demand."""

    min_replicas: int = Field(
        description="The minimum number of replicas to maintain, even when there is no load"
    )
    max_replicas: int = Field(
        description="The maximum number of replicas to scale up to under load"
    )


class EndpointPricing(TogetherJSONModel):
    """Pricing details for using an endpoint."""

    cents_per_minute: float = Field(
        description="Cost per minute of endpoint uptime in cents"
    )


class HardwareSpec(TogetherJSONModel):
    """Detailed specifications of a hardware configuration."""

    gpu_type: str = Field(description="The type/model of GPU")
    gpu_link: str = Field(description="The GPU interconnect technology")
    gpu_memory: Union[float, int] = Field(description="Amount of GPU memory in GB")
    gpu_count: int = Field(description="Number of GPUs in this configuration")


class HardwareAvailability(TogetherJSONModel):
    """Indicates the current availability status of a hardware configuration."""

    status: Literal["available", "unavailable", "insufficient"] = Field(
        description="The availability status of the hardware configuration"
    )


class HardwareWithStatus(TogetherJSONModel):
    """Hardware configuration details with optional availability status."""

    object: Literal["hardware"] = Field(description="The type of object")
    id: str = Field(description="Unique identifier for the hardware configuration")
    pricing: EndpointPricing = Field(
        description="Pricing details for this hardware configuration"
    )
    specs: HardwareSpec = Field(description="Detailed specifications of this hardware")
    availability: Optional[HardwareAvailability] = Field(
        default=None,
        description="Current availability status of this hardware configuration",
    )
    updated_at: datetime = Field(
        description="Timestamp of when the hardware status was last updated"
    )


class BaseEndpoint(TogetherJSONModel):
    """Base class for endpoint models with common fields."""

    object: Literal["endpoint"] = Field(description="The type of object")
    id: Optional[str] = Field(
        default=None, description="Unique identifier for the endpoint"
    )
    name: str = Field(description="System name for the endpoint")
    model: str = Field(description="The model deployed on this endpoint")
    type: str = Field(description="The type of endpoint")
    owner: str = Field(description="The owner of this endpoint")
    state: Literal[
        "PENDING", "STARTING", "STARTED", "STOPPING", "STOPPED", "FAILED", "ERROR"
    ] = Field(description="Current state of the endpoint")
    created_at: datetime = Field(description="Timestamp when the endpoint was created")


class ListEndpoint(BaseEndpoint):
    """Details about an endpoint when listed via the list endpoint."""

    type: Literal["dedicated", "serverless"] = Field(description="The type of endpoint")


class DedicatedEndpoint(BaseEndpoint):
    """Details about a dedicated endpoint deployment."""

    id: str = Field(description="Unique identifier for the endpoint")
    type: Literal["dedicated"] = Field(description="The type of endpoint")
    display_name: str = Field(description="Human-readable name for the endpoint")
    hardware: str = Field(
        description="The hardware configuration used for this endpoint"
    )
    autoscaling: Autoscaling = Field(
        description="Configuration for automatic scaling of the endpoint"
    )


__all__ = [
    "DedicatedEndpoint",
    "ListEndpoint",
    "Autoscaling",
    "EndpointPricing",
    "HardwareSpec",
    "HardwareAvailability",
    "HardwareWithStatus",
]



from __future__ import annotations

import os
from typing import Any

import click

import together
from together.cli.api.chat import chat, interactive
from together.cli.api.completions import completions
from together.cli.api.endpoints import endpoints
from together.cli.api.files import files
from together.cli.api.finetune import fine_tuning
from together.cli.api.images import images
from together.cli.api.models import models
from together.constants import MAX_RETRIES, TIMEOUT_SECS


def print_version(ctx: click.Context, params: Any, value: Any) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"Version {together.version}")
    ctx.exit()


@click.group()
@click.pass_context
@click.option(
    "--api-key",
    type=str,
    help="API Key. Defaults to environment variable `TOGETHER_API_KEY`",
    default=os.getenv("TOGETHER_API_KEY"),
)
@click.option(
    "--base-url", type=str, help="API Base URL. Defaults to Together AI endpoint."
)
@click.option(
    "--timeout", type=int, help=f"Request timeout. Defaults to {TIMEOUT_SECS} seconds"
)
@click.option(
    "--max-retries",
    type=int,
    help=f"Maximum number of HTTP retries. Defaults to {MAX_RETRIES}.",
)
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Print version",
)
@click.option("--debug", help="Debug mode", is_flag=True)
def main(
    ctx: click.Context,
    api_key: str | None,
    base_url: str | None,
    timeout: int | None,
    max_retries: int | None,
    debug: bool | None,
) -> None:
    """This is a sample CLI tool."""
    together.log = "debug" if debug else None
    ctx.obj = together.Together(
        api_key=api_key, base_url=base_url, timeout=timeout, max_retries=max_retries
    )


main.add_command(chat)
main.add_command(interactive)
main.add_command(completions)
main.add_command(images)
main.add_command(files)
main.add_command(fine_tuning)
main.add_command(models)
main.add_command(endpoints)

if __name__ == "__main__":
    main()



from __future__ import annotations

import warnings
from enum import Enum
from typing import Any, Dict, List

from pydantic import model_validator
from typing_extensions import Self

from together.types.abstract import BaseModel
from together.types.common import (
    DeltaContent,
    FinishReason,
    LogprobsPart,
    ObjectType,
    PromptPart,
    UsageData,
)


class MessageRole(str, Enum):
    ASSISTANT = "assistant"
    SYSTEM = "system"
    USER = "user"
    TOOL = "tool"


class ResponseFormatType(str, Enum):
    JSON_OBJECT = "json_object"
    JSON_SCHEMA = "json_schema"
    REGEX = "regex"


class FunctionCall(BaseModel):
    name: str | None = None
    arguments: str | None = None


class ToolCalls(BaseModel):
    id: str | None = None
    type: str | None = None
    function: FunctionCall | None = None


class ChatCompletionMessageContentType(str, Enum):
    TEXT = "text"
    IMAGE_URL = "image_url"
    VIDEO_URL = "video_url"


class ChatCompletionMessageContentImageURL(BaseModel):
    url: str


class ChatCompletionMessageContentVideoURL(BaseModel):
    url: str


class ChatCompletionMessageContent(BaseModel):
    type: ChatCompletionMessageContentType
    text: str | None = None
    image_url: ChatCompletionMessageContentImageURL | None = None
    video_url: ChatCompletionMessageContentVideoURL | None = None


class ChatCompletionMessage(BaseModel):
    role: MessageRole
    content: str | List[ChatCompletionMessageContent] | None = None
    tool_calls: List[ToolCalls] | None = None


class ResponseFormat(BaseModel):
    type: ResponseFormatType
    schema_: Dict[str, Any] | None = None
    pattern: str | None = None

    def to_dict(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {"type": self.type.value}
        if self.schema_ is not None:
            result["schema"] = self.schema_
        if self.pattern is not None:
            result["pattern"] = self.pattern
        return result


class FunctionTool(BaseModel):
    description: str | None = None
    name: str
    parameters: Dict[str, Any] | None = None


class FunctionToolChoice(BaseModel):
    name: str


class Tools(BaseModel):
    type: str
    function: FunctionTool


class ToolChoice(BaseModel):
    type: str
    function: FunctionToolChoice


class ToolChoiceEnum(str, Enum):
    Auto = "auto"
    Required = "required"


class ChatCompletionRequest(BaseModel):
    # list of messages
    messages: List[ChatCompletionMessage]
    # model name
    model: str
    # stopping criteria: max tokens to generate
    max_tokens: int | None = None
    # stopping criteria: list of strings to stop generation
    stop: List[str] | None = None
    # sampling hyperparameters
    temperature: float | None = None
    top_p: float | None = None
    top_k: int | None = None
    repetition_penalty: float | None = None
    presence_penalty: float | None = None
    frequency_penalty: float | None = None
    min_p: float | None = None
    logit_bias: Dict[str, float] | None = None
    seed: int | None = None
    # stream SSE token chunks
    stream: bool = False
    # return logprobs
    logprobs: int | None = None
    # echo prompt.
    # can be used with logprobs to return prompt logprobs
    echo: bool | None = None
    # number of output generations
    n: int | None = None
    # moderation model
    safety_model: str | None = None
    # constraints
    response_format: ResponseFormat | None = None
    tools: List[Tools] | None = None
    tool_choice: ToolChoice | ToolChoiceEnum | None = None

    # Raise warning if repetition_penalty is used with presence_penalty or frequency_penalty
    @model_validator(mode="after")
    def verify_parameters(self) -> Self:
        if self.repetition_penalty:
            if self.presence_penalty or self.frequency_penalty:
                warnings.warn(
                    "repetition_penalty is not advisable to be used alongside presence_penalty or frequency_penalty"
                )
        return self


class ChatCompletionChoicesData(BaseModel):
    index: int | None = None
    logprobs: LogprobsPart | None = None
    seed: int | None = None
    finish_reason: FinishReason | None = None
    message: ChatCompletionMessage | None = None


class ChatCompletionResponse(BaseModel):
    # request id
    id: str | None = None
    # object type
    object: ObjectType | None = None
    # created timestamp
    created: int | None = None
    # model name
    model: str | None = None
    # choices list
    choices: List[ChatCompletionChoicesData] | None = None
    # prompt list
    prompt: List[PromptPart] | List[None] | None = None
    # token usage data
    usage: UsageData | None = None


class ChatCompletionChoicesChunk(BaseModel):
    index: int | None = None
    logprobs: float | None = None
    seed: int | None = None
    finish_reason: FinishReason | None = None
    delta: DeltaContent | None = None


class ChatCompletionChunk(BaseModel):
    # request id
    id: str | None = None
    # object type
    object: ObjectType | None = None
    # created timestamp
    created: int | None = None
    # model name
    model: str | None = None
    # delta content
    choices: List[ChatCompletionChoicesChunk] | None = None
    # finish reason
    finish_reason: FinishReason | None = None
    # token usage data
    usage: UsageData | None = None



from __future__ import annotations

import re
from pathlib import Path
from typing import List, Dict, Literal

from rich import print as rprint

from together.abstract import api_requestor
from together.filemanager import DownloadManager
from together.together_response import TogetherResponse
from together.types import (
    CosineLRScheduler,
    CosineLRSchedulerArgs,
    FinetuneCheckpoint,
    FinetuneDownloadResult,
    FinetuneList,
    FinetuneListEvents,
    FinetuneLRScheduler,
    FinetuneRequest,
    FinetuneResponse,
    FinetuneTrainingLimits,
    FullTrainingType,
    LinearLRScheduler,
    LinearLRSchedulerArgs,
    LoRATrainingType,
    TogetherClient,
    TogetherRequest,
    TrainingMethodDPO,
    TrainingMethodSFT,
    TrainingType,
)
from together.types.finetune import DownloadCheckpointType
from together.utils import log_warn_once, normalize_key


_FT_JOB_WITH_STEP_REGEX = r"^ft-[\dabcdef-]+:\d+$"


AVAILABLE_TRAINING_METHODS = {
    TrainingMethodSFT().method,
    TrainingMethodDPO().method,
}


def create_finetune_request(
    model_limits: FinetuneTrainingLimits,
    training_file: str,
    model: str | None = None,
    n_epochs: int = 1,
    validation_file: str | None = "",
    n_evals: int | None = 0,
    n_checkpoints: int | None = 1,
    batch_size: int | Literal["max"] = "max",
    learning_rate: float | None = 0.00001,
    lr_scheduler_type: Literal["linear", "cosine"] = "cosine",
    min_lr_ratio: float = 0.0,
    scheduler_num_cycles: float = 0.5,
    warmup_ratio: float | None = None,
    max_grad_norm: float = 1.0,
    weight_decay: float = 0.0,
    lora: bool = False,
    lora_r: int | None = None,
    lora_dropout: float | None = 0,
    lora_alpha: float | None = None,
    lora_trainable_modules: str | None = "all-linear",
    suffix: str | None = None,
    wandb_api_key: str | None = None,
    wandb_base_url: str | None = None,
    wandb_project_name: str | None = None,
    wandb_name: str | None = None,
    train_on_inputs: bool | Literal["auto"] | None = None,
    training_method: str = "sft",
    dpo_beta: float | None = None,
    dpo_normalize_logratios_by_length: bool = False,
    rpo_alpha: float | None = None,
    simpo_gamma: float | None = None,
    from_checkpoint: str | None = None,
    hf_api_token: str | None = None,
    hf_output_repo_name: str | None = None,
) -> FinetuneRequest:
    if model is not None and from_checkpoint is not None:
        raise ValueError(
            "You must specify either a model or a checkpoint to start a job from, not both"
        )

    if model is None and from_checkpoint is None:
        raise ValueError("You must specify either a model or a checkpoint")

    model_or_checkpoint = model or from_checkpoint

    if batch_size == "max":
        log_warn_once(
            "Starting from together>=1.3.0, "
            "the default batch size is set to the maximum allowed value for each model."
        )
    if warmup_ratio is None:
        warmup_ratio = 0.0

    training_type: TrainingType = FullTrainingType()
    max_batch_size: int = 0
    max_batch_size_dpo: int = 0
    min_batch_size: int = 0
    if lora:
        if model_limits.lora_training is None:
            raise ValueError(
                f"LoRA adapters are not supported for the selected model ({model_or_checkpoint})."
            )

        if lora_dropout is not None:
            if not 0 <= lora_dropout < 1.0:
                raise ValueError("LoRA dropout must be in [0, 1) range.")

        lora_r = lora_r if lora_r is not None else model_limits.lora_training.max_rank
        lora_alpha = lora_alpha if lora_alpha is not None else lora_r * 2
        training_type = LoRATrainingType(
            lora_r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            lora_trainable_modules=lora_trainable_modules,
        )

        max_batch_size = model_limits.lora_training.max_batch_size
        min_batch_size = model_limits.lora_training.min_batch_size
        max_batch_size_dpo = model_limits.lora_training.max_batch_size_dpo
    else:
        if model_limits.full_training is None:
            raise ValueError(
                f"Full training is not supported for the selected model ({model_or_checkpoint})."
            )

        max_batch_size = model_limits.full_training.max_batch_size
        min_batch_size = model_limits.full_training.min_batch_size
        max_batch_size_dpo = model_limits.full_training.max_batch_size_dpo

    if batch_size == "max":
        if training_method == "dpo":
            batch_size = max_batch_size_dpo
        else:
            batch_size = max_batch_size

    if training_method == "sft":
        if batch_size > max_batch_size:
            raise ValueError(
                f"Requested batch size of {batch_size} is higher that the maximum allowed value of {max_batch_size}."
            )
    elif training_method == "dpo":
        if batch_size > max_batch_size_dpo:
            raise ValueError(
                f"Requested batch size of {batch_size} is higher that the maximum allowed value of {max_batch_size_dpo}."
            )

    if batch_size < min_batch_size:
        raise ValueError(
            f"Requested batch size of {batch_size} is lower that the minimum allowed value of {min_batch_size}."
        )

    if warmup_ratio > 1 or warmup_ratio < 0:
        raise ValueError(f"Warmup ratio should be between 0 and 1 (got {warmup_ratio})")

    if min_lr_ratio is not None and (min_lr_ratio > 1 or min_lr_ratio < 0):
        raise ValueError(
            f"Min learning rate ratio should be between 0 and 1 (got {min_lr_ratio})"
        )

    if max_grad_norm < 0:
        raise ValueError(
            f"Max gradient norm should be non-negative (got {max_grad_norm})"
        )

    if weight_decay is not None and (weight_decay < 0):
        raise ValueError(f"Weight decay should be non-negative (got {weight_decay})")

    if training_method not in AVAILABLE_TRAINING_METHODS:
        raise ValueError(
            f"training_method must be one of {', '.join(AVAILABLE_TRAINING_METHODS)}"
        )

    if train_on_inputs is not None and training_method != "sft":
        raise ValueError("train_on_inputs is only supported for SFT training")

    if train_on_inputs is None and training_method == "sft":
        log_warn_once(
            "train_on_inputs is not set for SFT training, it will be set to 'auto'"
        )
        train_on_inputs = "auto"

    if dpo_beta is not None and training_method != "dpo":
        raise ValueError("dpo_beta is only supported for DPO training")
    if dpo_normalize_logratios_by_length and training_method != "dpo":
        raise ValueError(
            "dpo_normalize_logratios_by_length=True is only supported for DPO training"
        )
    if rpo_alpha is not None:
        if training_method != "dpo":
            raise ValueError("rpo_alpha is only supported for DPO training")
        if not rpo_alpha >= 0.0:
            raise ValueError(f"rpo_alpha should be non-negative (got {rpo_alpha})")

    if simpo_gamma is not None:
        if training_method != "dpo":
            raise ValueError("simpo_gamma is only supported for DPO training")
        if not simpo_gamma >= 0.0:
            raise ValueError(f"simpo_gamma should be non-negative (got {simpo_gamma})")

    lr_scheduler: FinetuneLRScheduler
    if lr_scheduler_type == "cosine":
        if scheduler_num_cycles <= 0.0:
            raise ValueError(
                f"Number of cycles should be greater than 0 (got {scheduler_num_cycles})"
            )

        lr_scheduler = CosineLRScheduler(
            lr_scheduler_args=CosineLRSchedulerArgs(
                min_lr_ratio=min_lr_ratio, num_cycles=scheduler_num_cycles
            ),
        )
    else:
        lr_scheduler = LinearLRScheduler(
            lr_scheduler_args=LinearLRSchedulerArgs(min_lr_ratio=min_lr_ratio),
        )

    training_method_cls: TrainingMethodSFT | TrainingMethodDPO
    if training_method == "sft":
        training_method_cls = TrainingMethodSFT(train_on_inputs=train_on_inputs)
    elif training_method == "dpo":
        if simpo_gamma is not None and simpo_gamma > 0:
            dpo_reference_free = True
            dpo_normalize_logratios_by_length = True
            rprint(
                f"Parameter simpo_gamma was set to {simpo_gamma}. "
                "SimPO training detected. Reference logits will not be used "
                "and length normalization of log-probabilities will be enabled."
            )
        else:
            dpo_reference_free = False

        training_method_cls = TrainingMethodDPO(
            dpo_beta=dpo_beta,
            dpo_normalize_logratios_by_length=dpo_normalize_logratios_by_length,
            dpo_reference_free=dpo_reference_free,
            rpo_alpha=rpo_alpha,
            simpo_gamma=simpo_gamma,
        )

    finetune_request = FinetuneRequest(
        model=model,
        training_file=training_file,
        validation_file=validation_file,
        n_epochs=n_epochs,
        n_evals=n_evals,
        n_checkpoints=n_checkpoints,
        batch_size=batch_size,
        learning_rate=learning_rate,
        lr_scheduler=lr_scheduler,
        warmup_ratio=warmup_ratio,
        max_grad_norm=max_grad_norm,
        weight_decay=weight_decay,
        training_type=training_type,
        suffix=suffix,
        wandb_key=wandb_api_key,
        wandb_base_url=wandb_base_url,
        wandb_project_name=wandb_project_name,
        wandb_name=wandb_name,
        training_method=training_method_cls,
        from_checkpoint=from_checkpoint,
        hf_api_token=hf_api_token,
        hf_output_repo_name=hf_output_repo_name,
    )

    return finetune_request


def _parse_raw_checkpoints(
    checkpoints: List[Dict[str, str]], id: str
) -> List[FinetuneCheckpoint]:
    """
    Helper function to process raw checkpoints and create checkpoint list.

    Args:
        checkpoints (List[Dict[str, str]]): List of raw checkpoints metadata
        id (str): Fine-tune job ID

    Returns:
        List[FinetuneCheckpoint]: List of available checkpoints
    """

    parsed_checkpoints = []
    for checkpoint in checkpoints:
        step = checkpoint["step"]
        checkpoint_type = checkpoint["checkpoint_type"]
        checkpoint_name = (
            f"{id}:{step}" if "intermediate" in checkpoint_type.lower() else id
        )

        parsed_checkpoints.append(
            FinetuneCheckpoint(
                type=checkpoint_type,
                timestamp=checkpoint["created_at"],
                name=checkpoint_name,
            )
        )

    parsed_checkpoints.sort(key=lambda x: x.timestamp, reverse=True)
    return parsed_checkpoints


class FineTuning:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    def create(
        self,
        *,
        training_file: str,
        model: str | None = None,
        n_epochs: int = 1,
        validation_file: str | None = "",
        n_evals: int | None = 0,
        n_checkpoints: int | None = 1,
        batch_size: int | Literal["max"] = "max",
        learning_rate: float | None = 0.00001,
        lr_scheduler_type: Literal["linear", "cosine"] = "cosine",
        min_lr_ratio: float = 0.0,
        scheduler_num_cycles: float = 0.5,
        warmup_ratio: float = 0.0,
        max_grad_norm: float = 1.0,
        weight_decay: float = 0.0,
        lora: bool = True,
        lora_r: int | None = None,
        lora_dropout: float | None = 0,
        lora_alpha: float | None = None,
        lora_trainable_modules: str | None = "all-linear",
        suffix: str | None = None,
        wandb_api_key: str | None = None,
        wandb_base_url: str | None = None,
        wandb_project_name: str | None = None,
        wandb_name: str | None = None,
        verbose: bool = False,
        model_limits: FinetuneTrainingLimits | None = None,
        train_on_inputs: bool | Literal["auto"] | None = None,
        training_method: str = "sft",
        dpo_beta: float | None = None,
        dpo_normalize_logratios_by_length: bool = False,
        rpo_alpha: float | None = None,
        simpo_gamma: float | None = None,
        from_checkpoint: str | None = None,
        hf_api_token: str | None = None,
        hf_output_repo_name: str | None = None,
    ) -> FinetuneResponse:
        """
        Method to initiate a fine-tuning job

        Args:
            training_file (str): File-ID of a file uploaded to the Together API
            model (str, optional): Name of the base model to run fine-tune job on
            n_epochs (int, optional): Number of epochs for fine-tuning. Defaults to 1.
            validation file (str, optional): File ID of a file uploaded to the Together API for validation.
            n_evals (int, optional): Number of evaluation loops to run. Defaults to 0.
            n_checkpoints (int, optional): Number of checkpoints to save during fine-tuning.
                Defaults to 1.
            batch_size (int or "max"): Batch size for fine-tuning. Defaults to max.
            learning_rate (float, optional): Learning rate multiplier to use for training
                Defaults to 0.00001.
            lr_scheduler_type (Literal["linear", "cosine"]): Learning rate scheduler type. Defaults to "cosine".
            min_lr_ratio (float, optional): Min learning rate ratio of the initial learning rate for
                the learning rate scheduler. Defaults to 0.0.
            scheduler_num_cycles (float, optional): Number or fraction of cycles for the cosine learning rate scheduler. Defaults to 0.5.
            warmup_ratio (float, optional): Warmup ratio for the learning rate scheduler.
            max_grad_norm (float, optional): Max gradient norm. Defaults to 1.0, set to 0 to disable.
            weight_decay (float, optional): Weight decay. Defaults to 0.0.
            lora (bool, optional): Whether to use LoRA adapters. Defaults to True.
            lora_r (int, optional): Rank of LoRA adapters. Defaults to 8.
            lora_dropout (float, optional): Dropout rate for LoRA adapters. Defaults to 0.
            lora_alpha (float, optional): Alpha for LoRA adapters. Defaults to 8.
            lora_trainable_modules (str, optional): Trainable modules for LoRA adapters. Defaults to "all-linear".
            suffix (str, optional): Up to 40 character suffix that will be added to your fine-tuned model name.
                Defaults to None.
            wandb_api_key (str, optional): API key for Weights & Biases integration.
                Defaults to None.
            wandb_base_url (str, optional): Base URL for Weights & Biases integration.
                Defaults to None.
            wandb_project_name (str, optional): Project name for Weights & Biases integration.
                Defaults to None.
            wandb_name (str, optional): Run name for Weights & Biases integration.
                Defaults to None.
            verbose (bool, optional): whether to print the job parameters before submitting a request.
                Defaults to False.
            model_limits (FinetuneTrainingLimits, optional): Limits for the hyperparameters the model in Fine-tuning.
                Defaults to None.
            train_on_inputs (bool or "auto", optional): Whether to mask the user messages in conversational data or prompts in instruction data.
                "auto" will automatically determine whether to mask the inputs based on the data format.
                For datasets with the "text" field (general format), inputs will not be masked.
                For datasets with the "messages" field (conversational format) or "prompt" and "completion" fields
                (Instruction format), inputs will be masked.
                Defaults to None, or "auto" if training_method is "sft" (set in create_finetune_request).
            training_method (str, optional): Training method. Defaults to "sft".
                Supported methods: "sft", "dpo".
            dpo_beta (float, optional): DPO beta parameter. Defaults to None.
            dpo_normalize_logratios_by_length (bool): Whether or not normalize logratios by sample length. Defaults to False,
            rpo_alpha (float, optional): RPO alpha parameter of DPO training to include NLL in the loss. Defaults to None.
            simpo_gamma: (float, optional): SimPO gamma parameter. Defaults to None.
            from_checkpoint (str, optional): The checkpoint identifier to continue training from a previous fine-tuning job.
                The format: {$JOB_ID/$OUTPUT_MODEL_NAME}:{$STEP}.
                The step value is optional, without it the final checkpoint will be used.
            hf_api_token (str, optional): API key for the Hugging Face Hub. Defaults to None.
            hf_output_repo_name (str, optional): HF repo to upload the fine-tuned model to. Defaults to None.

        Returns:
            FinetuneResponse: Object containing information about fine-tuning job.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        if model_limits is None:
            # mypy doesn't understand that model or from_checkpoint is not None
            if model is not None:
                model_name = model
            elif from_checkpoint is not None:
                model_name = from_checkpoint.split(":")[0]
            else:
                # this branch is unreachable, but mypy doesn't know that
                pass
            model_limits = self.get_model_limits(model=model_name)

        finetune_request = create_finetune_request(
            model_limits=model_limits,
            training_file=training_file,
            model=model,
            n_epochs=n_epochs,
            validation_file=validation_file,
            n_evals=n_evals,
            n_checkpoints=n_checkpoints,
            batch_size=batch_size,
            learning_rate=learning_rate,
            lr_scheduler_type=lr_scheduler_type,
            min_lr_ratio=min_lr_ratio,
            scheduler_num_cycles=scheduler_num_cycles,
            warmup_ratio=warmup_ratio,
            max_grad_norm=max_grad_norm,
            weight_decay=weight_decay,
            lora=lora,
            lora_r=lora_r,
            lora_dropout=lora_dropout,
            lora_alpha=lora_alpha,
            lora_trainable_modules=lora_trainable_modules,
            suffix=suffix,
            wandb_api_key=wandb_api_key,
            wandb_base_url=wandb_base_url,
            wandb_project_name=wandb_project_name,
            wandb_name=wandb_name,
            train_on_inputs=train_on_inputs,
            training_method=training_method,
            dpo_beta=dpo_beta,
            dpo_normalize_logratios_by_length=dpo_normalize_logratios_by_length,
            rpo_alpha=rpo_alpha,
            simpo_gamma=simpo_gamma,
            from_checkpoint=from_checkpoint,
            hf_api_token=hf_api_token,
            hf_output_repo_name=hf_output_repo_name,
        )

        if verbose:
            rprint(
                "Submitting a fine-tuning job with the following parameters:",
                finetune_request,
            )
        parameter_payload = finetune_request.model_dump(exclude_none=True)

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="POST",
                url="fine-tunes",
                params=parameter_payload,
            ),
            stream=False,
        )
        assert isinstance(response, TogetherResponse)

        return FinetuneResponse(**response.data)

    def list(self) -> FinetuneList:
        """
        Lists fine-tune job history

        Returns:
            FinetuneList: Object containing a list of fine-tune jobs
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url="fine-tunes",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneList(**response.data)

    def retrieve(self, id: str) -> FinetuneResponse:
        """
        Retrieves fine-tune job details

        Args:
            id (str): Fine-tune ID to retrieve. A string that starts with `ft-`.

        Returns:
            FinetuneResponse: Object containing information about fine-tuning job.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url=f"fine-tunes/{id}",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneResponse(**response.data)

    def cancel(self, id: str) -> FinetuneResponse:
        """
        Method to cancel a running fine-tuning job

        Args:
            id (str): Fine-tune ID to cancel. A string that starts with `ft-`.

        Returns:
            FinetuneResponse: Object containing information about cancelled fine-tuning job.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="POST",
                url=f"fine-tunes/{id}/cancel",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneResponse(**response.data)

    def list_events(self, id: str) -> FinetuneListEvents:
        """
        Lists events of a fine-tune job

        Args:
            id (str): Fine-tune ID to list events for. A string that starts with `ft-`.

        Returns:
            FinetuneListEvents: Object containing list of fine-tune events
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url=f"fine-tunes/{id}/events",
            ),
            stream=False,
        )
        assert isinstance(response, TogetherResponse)

        return FinetuneListEvents(**response.data)

    def list_checkpoints(self, id: str) -> List[FinetuneCheckpoint]:
        """
        List available checkpoints for a fine-tuning job

        Args:
            id (str): Unique identifier of the fine-tune job to list checkpoints for

        Returns:
            List[FinetuneCheckpoint]: List of available checkpoints
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url=f"fine-tunes/{id}/checkpoints",
            ),
            stream=False,
        )
        assert isinstance(response, TogetherResponse)

        raw_checkpoints = response.data["data"]
        return _parse_raw_checkpoints(raw_checkpoints, id)

    def download(
        self,
        id: str,
        *,
        output: Path | str | None = None,
        checkpoint_step: int | None = None,
        checkpoint_type: DownloadCheckpointType | str = DownloadCheckpointType.DEFAULT,
    ) -> FinetuneDownloadResult:
        """
        Downloads compressed fine-tuned model or checkpoint to local disk.

        Defaults file location to `$PWD/{model_name}.{extension}`

        Args:
            id (str): Fine-tune ID to download. A string that starts with `ft-`.
            output (pathlib.Path | str, optional): Specifies output file name for downloaded model.
                Defaults to None.
            checkpoint_step (int, optional): Specifies step number for checkpoint to download.
                Defaults to -1 (download the final model)
            checkpoint_type (CheckpointType | str, optional): Specifies which checkpoint to download.
                Defaults to CheckpointType.DEFAULT.

        Returns:
            FinetuneDownloadResult: Object containing downloaded model metadata
        """

        if re.match(_FT_JOB_WITH_STEP_REGEX, id) is not None:
            if checkpoint_step is None:
                checkpoint_step = int(id.split(":")[1])
                id = id.split(":")[0]
            else:
                raise ValueError(
                    "Fine-tuning job ID {id} contains a colon to specify the step to download, but `checkpoint_step` "
                    "was also set. Remove one of the step specifiers to proceed."
                )

        url = f"finetune/download?ft_id={id}"

        if checkpoint_step is not None:
            url += f"&checkpoint_step={checkpoint_step}"

        ft_job = self.retrieve(id)

        # convert str to DownloadCheckpointType
        if isinstance(checkpoint_type, str):
            try:
                checkpoint_type = DownloadCheckpointType(checkpoint_type.lower())
            except ValueError:
                enum_strs = ", ".join(e.value for e in DownloadCheckpointType)
                raise ValueError(
                    f"Invalid checkpoint type: {checkpoint_type}. Choose one of {{{enum_strs}}}."
                )

        if isinstance(ft_job.training_type, FullTrainingType):
            if checkpoint_type != DownloadCheckpointType.DEFAULT:
                raise ValueError(
                    "Only DEFAULT checkpoint type is allowed for FullTrainingType"
                )
            url += "&checkpoint=model_output_path"
        elif isinstance(ft_job.training_type, LoRATrainingType):
            if checkpoint_type == DownloadCheckpointType.DEFAULT:
                checkpoint_type = DownloadCheckpointType.MERGED

            if checkpoint_type in {
                DownloadCheckpointType.MERGED,
                DownloadCheckpointType.ADAPTER,
            }:
                url += f"&checkpoint={checkpoint_type.value}"
            else:
                raise ValueError(
                    f"Invalid checkpoint type for LoRATrainingType: {checkpoint_type}"
                )

        remote_name = ft_job.output_name

        download_manager = DownloadManager(self._client)

        if isinstance(output, str):
            output = Path(output)

        downloaded_filename, file_size = download_manager.download(
            url, output, normalize_key(remote_name or id), fetch_metadata=True
        )

        return FinetuneDownloadResult(
            object="local",
            id=id,
            checkpoint_step=checkpoint_step,
            filename=downloaded_filename,
            size=file_size,
        )

    def get_model_limits(self, *, model: str) -> FinetuneTrainingLimits:
        """
        Requests training limits for a specific model

        Args:
            model_name (str): Name of the model to get limits for

        Returns:
            FinetuneTrainingLimits: Object containing training limits for the model
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        model_limits_response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url="fine-tunes/models/limits",
                params={"model_name": model},
            ),
            stream=False,
        )

        model_limits = FinetuneTrainingLimits(**model_limits_response.data)

        return model_limits


class AsyncFineTuning:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    async def create(
        self,
        *,
        training_file: str,
        model: str | None = None,
        n_epochs: int = 1,
        validation_file: str | None = "",
        n_evals: int | None = 0,
        n_checkpoints: int | None = 1,
        batch_size: int | Literal["max"] = "max",
        learning_rate: float | None = 0.00001,
        lr_scheduler_type: Literal["linear", "cosine"] = "cosine",
        min_lr_ratio: float = 0.0,
        scheduler_num_cycles: float = 0.5,
        warmup_ratio: float = 0.0,
        max_grad_norm: float = 1.0,
        weight_decay: float = 0.0,
        lora: bool = True,
        lora_r: int | None = None,
        lora_dropout: float | None = 0,
        lora_alpha: float | None = None,
        lora_trainable_modules: str | None = "all-linear",
        suffix: str | None = None,
        wandb_api_key: str | None = None,
        wandb_base_url: str | None = None,
        wandb_project_name: str | None = None,
        wandb_name: str | None = None,
        verbose: bool = False,
        model_limits: FinetuneTrainingLimits | None = None,
        train_on_inputs: bool | Literal["auto"] | None = None,
        training_method: str = "sft",
        dpo_beta: float | None = None,
        dpo_normalize_logratios_by_length: bool = False,
        rpo_alpha: float | None = None,
        simpo_gamma: float | None = None,
        from_checkpoint: str | None = None,
        hf_api_token: str | None = None,
        hf_output_repo_name: str | None = None,
    ) -> FinetuneResponse:
        """
        Async method to initiate a fine-tuning job

        Args:
            training_file (str): File-ID of a file uploaded to the Together API
            model (str, optional): Name of the base model to run fine-tune job on
            n_epochs (int, optional): Number of epochs for fine-tuning. Defaults to 1.
            validation file (str, optional): File ID of a file uploaded to the Together API for validation.
            n_evals (int, optional): Number of evaluation loops to run. Defaults to 0.
            n_checkpoints (int, optional): Number of checkpoints to save during fine-tuning.
                Defaults to 1.
            batch_size (int, optional): Batch size for fine-tuning. Defaults to max.
            learning_rate (float, optional): Learning rate multiplier to use for training
                Defaults to 0.00001.
            lr_scheduler_type (Literal["linear", "cosine"]): Learning rate scheduler type. Defaults to "cosine".
            min_lr_ratio (float, optional): Min learning rate ratio of the initial learning rate for
                the learning rate scheduler. Defaults to 0.0.
            scheduler_num_cycles (float, optional): Number or fraction of cycles for the cosine learning rate scheduler. Defaults to 0.5.
            warmup_ratio (float, optional): Warmup ratio for the learning rate scheduler.
            max_grad_norm (float, optional): Max gradient norm. Defaults to 1.0, set to 0 to disable.
            weight_decay (float, optional): Weight decay. Defaults to 0.0.
            lora (bool, optional): Whether to use LoRA adapters. Defaults to True.
            lora_r (int, optional): Rank of LoRA adapters. Defaults to 8.
            lora_dropout (float, optional): Dropout rate for LoRA adapters. Defaults to 0.
            lora_alpha (float, optional): Alpha for LoRA adapters. Defaults to 8.
            lora_trainable_modules (str, optional): Trainable modules for LoRA adapters. Defaults to "all-linear".
            suffix (str, optional): Up to 40 character suffix that will be added to your fine-tuned model name.
                Defaults to None.
            wandb_api_key (str, optional): API key for Weights & Biases integration.
                Defaults to None.
            wandb_base_url (str, optional): Base URL for Weights & Biases integration.
                Defaults to None.
            wandb_project_name (str, optional): Project name for Weights & Biases integration.
                Defaults to None.
            wandb_name (str, optional): Run name for Weights & Biases integration.
                Defaults to None.
            verbose (bool, optional): whether to print the job parameters before submitting a request.
                Defaults to False.
            model_limits (FinetuneTrainingLimits, optional): Limits for the hyperparameters the model in Fine-tuning.
                Defaults to None.
            train_on_inputs (bool or "auto"): Whether to mask the user messages in conversational data or prompts in instruction data.
                "auto" will automatically determine whether to mask the inputs based on the data format.
                For datasets with the "text" field (general format), inputs will not be masked.
                For datasets with the "messages" field (conversational format) or "prompt" and "completion" fields
                (Instruction format), inputs will be masked.
                Defaults to None, or "auto" if training_method is "sft" (set in create_finetune_request).
            training_method (str, optional): Training method. Defaults to "sft".
                Supported methods: "sft", "dpo".
            dpo_beta (float, optional): DPO beta parameter. Defaults to None.
            dpo_normalize_logratios_by_length (bool): Whether or not normalize logratios by sample length. Defaults to False,
            rpo_alpha (float, optional): RPO alpha parameter of DPO training to include NLL in the loss. Defaults to None.
            simpo_gamma: (float, optional): SimPO gamma parameter. Defaults to None.
            from_checkpoint (str, optional): The checkpoint identifier to continue training from a previous fine-tuning job.
                The format: {$JOB_ID/$OUTPUT_MODEL_NAME}:{$STEP}.
                The step value is optional, without it the final checkpoint will be used.
            hf_api_token (str, optional): API key for the Huggging Face Hub. Defaults to None.
            hf_output_repo_name (str, optional): HF repo to upload the fine-tuned model to. Defaults to None.

        Returns:
            FinetuneResponse: Object containing information about fine-tuning job.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        if model_limits is None:
            # mypy doesn't understand that model or from_checkpoint is not None
            if model is not None:
                model_name = model
            elif from_checkpoint is not None:
                model_name = from_checkpoint.split(":")[0]
            else:
                # this branch is unreachable, but mypy doesn't know that
                pass
            model_limits = await self.get_model_limits(model=model_name)

        finetune_request = create_finetune_request(
            model_limits=model_limits,
            training_file=training_file,
            model=model,
            n_epochs=n_epochs,
            validation_file=validation_file,
            n_evals=n_evals,
            n_checkpoints=n_checkpoints,
            batch_size=batch_size,
            learning_rate=learning_rate,
            lr_scheduler_type=lr_scheduler_type,
            min_lr_ratio=min_lr_ratio,
            scheduler_num_cycles=scheduler_num_cycles,
            warmup_ratio=warmup_ratio,
            max_grad_norm=max_grad_norm,
            weight_decay=weight_decay,
            lora=lora,
            lora_r=lora_r,
            lora_dropout=lora_dropout,
            lora_alpha=lora_alpha,
            lora_trainable_modules=lora_trainable_modules,
            suffix=suffix,
            wandb_api_key=wandb_api_key,
            wandb_base_url=wandb_base_url,
            wandb_project_name=wandb_project_name,
            wandb_name=wandb_name,
            train_on_inputs=train_on_inputs,
            training_method=training_method,
            dpo_beta=dpo_beta,
            dpo_normalize_logratios_by_length=dpo_normalize_logratios_by_length,
            rpo_alpha=rpo_alpha,
            simpo_gamma=simpo_gamma,
            from_checkpoint=from_checkpoint,
            hf_api_token=hf_api_token,
            hf_output_repo_name=hf_output_repo_name,
        )

        if verbose:
            rprint(
                "Submitting a fine-tuning job with the following parameters:",
                finetune_request,
            )
        parameter_payload = finetune_request.model_dump(exclude_none=True)

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="POST",
                url="fine-tunes",
                params=parameter_payload,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneResponse(**response.data)

    async def list(self) -> FinetuneList:
        """
        Async method to list fine-tune job history

        Returns:
            FinetuneList: Object containing a list of fine-tune jobs
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url="fine-tunes",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneList(**response.data)

    async def retrieve(self, id: str) -> FinetuneResponse:
        """
        Async method to retrieve fine-tune job details

        Args:
            id (str): Fine-tune ID to retrieve. A string that starts with `ft-`.

        Returns:
            FinetuneResponse: Object containing information about fine-tuning job.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url=f"fine-tunes/{id}",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneResponse(**response.data)

    async def cancel(self, id: str) -> FinetuneResponse:
        """
        Async method to cancel a running fine-tuning job

        Args:
            id (str): Fine-tune ID to cancel. A string that starts with `ft-`.

        Returns:
            FinetuneResponse: Object containing information about cancelled fine-tuning job.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="POST",
                url=f"fine-tunes/{id}/cancel",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return FinetuneResponse(**response.data)

    async def list_events(self, id: str) -> FinetuneListEvents:
        """
        List fine-tuning events

        Args:
            id (str): Unique identifier of the fine-tune job to list events for

        Returns:
            FinetuneListEvents: Object containing list of fine-tune job events
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        events_response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url=f"fine-tunes/{normalize_key(id)}/events",
            ),
            stream=False,
        )
        assert isinstance(events_response, TogetherResponse)

        return FinetuneListEvents(**events_response.data)

    async def list_checkpoints(self, id: str) -> List[FinetuneCheckpoint]:
        """
        List available checkpoints for a fine-tuning job

        Args:
            id (str): Unique identifier of the fine-tune job to list checkpoints for

        Returns:
            List[FinetuneCheckpoint]: List of available checkpoints
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url=f"fine-tunes/{id}/checkpoints",
            ),
            stream=False,
        )
        assert isinstance(response, TogetherResponse)

        raw_checkpoints = response.data["data"]
        return _parse_raw_checkpoints(raw_checkpoints, id)

    async def download(
        self, id: str, *, output: str | None = None, checkpoint_step: int = -1
    ) -> str:
        """
        TODO: Implement async download method
        """

        raise NotImplementedError(
            "AsyncFineTuning.download not implemented. "
            "Please use FineTuning.download function instead."
        )

    async def get_model_limits(self, *, model: str) -> FinetuneTrainingLimits:
        """
        Requests training limits for a specific model

        Args:
            model_name (str): Name of the model to get limits for

        Returns:
            FinetuneTrainingLimits: Object containing training limits for the model
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        model_limits_response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url="fine-tunes/models/limits",
                params={"model": model},
            ),
            stream=False,
        )

        model_limits = FinetuneTrainingLimits(**model_limits_response.data)

        return model_limits



from __future__ import annotations

import json
import os
from pathlib import Path
from traceback import format_exc
from typing import Any, Dict, List


from together.constants import (
    MAX_FILE_SIZE_GB,
    MIN_SAMPLES,
    NUM_BYTES_IN_GB,
    PARQUET_EXPECTED_COLUMNS,
    JSONL_REQUIRED_COLUMNS_MAP,
    REQUIRED_COLUMNS_MESSAGE,
    POSSIBLE_ROLES_CONVERSATION,
    DatasetFormat,
)


class InvalidFileFormatError(ValueError):
    """Exception raised for invalid file formats during file checks."""

    def __init__(
        self,
        message: str = "",
        line_number: int | None = None,
        error_source: str | None = None,
    ) -> None:
        super().__init__(message)
        self.message = message
        self.line_number = line_number
        self.error_source = error_source


def check_file(
    file: Path | str,
) -> Dict[str, Any]:
    if not isinstance(file, Path):
        file = Path(file)

    report_dict = {
        "is_check_passed": True,
        "message": "Checks passed",
        "found": None,
        "file_size": None,
        "utf8": None,
        "line_type": None,
        "text_field": None,
        "key_value": None,
        "has_min_samples": None,
        "num_samples": None,
        "load_json": None,
    }

    if not file.is_file():
        report_dict["found"] = False
        report_dict["is_check_passed"] = False
        return report_dict
    else:
        report_dict["found"] = True

    file_size = os.stat(file.as_posix()).st_size

    if file_size > MAX_FILE_SIZE_GB * NUM_BYTES_IN_GB:
        report_dict["message"] = (
            f"Maximum supported file size is {MAX_FILE_SIZE_GB} GB. Found file with size of {round(file_size / NUM_BYTES_IN_GB ,3)} GB."
        )
        report_dict["is_check_passed"] = False
    elif file_size == 0:
        report_dict["message"] = "File is empty"
        report_dict["file_size"] = 0
        report_dict["is_check_passed"] = False
        return report_dict
    else:
        report_dict["file_size"] = file_size

    data_report_dict = {}
    if file.suffix == ".jsonl":
        report_dict["filetype"] = "jsonl"
        data_report_dict = _check_jsonl(file)
    elif file.suffix == ".parquet":
        report_dict["filetype"] = "parquet"
        data_report_dict = _check_parquet(file)
    else:
        report_dict["filetype"] = (
            f"Unknown extension of file {file}. "
            "Only files with extensions .jsonl and .parquet are supported."
        )
        report_dict["is_check_passed"] = False

    report_dict.update(data_report_dict)

    return report_dict


def validate_messages(messages: List[Dict[str, str | bool]], idx: int) -> None:
    """Validate the messages column."""
    if not isinstance(messages, list):
        raise InvalidFileFormatError(
            message=f"Invalid format on line {idx + 1} of the input file. "
            f"Expected a list of messages. Found {type(messages)}",
            line_number=idx + 1,
            error_source="key_value",
        )
    if not messages:
        raise InvalidFileFormatError(
            message=f"Invalid format on line {idx + 1} of the input file. "
            f"Expected a non-empty list of messages. Found empty list",
            line_number=idx + 1,
            error_source="key_value",
        )

    has_weights = any("weight" in message for message in messages)

    previous_role = None
    for message in messages:
        if not isinstance(message, dict):
            raise InvalidFileFormatError(
                message=f"Invalid format on line {idx + 1} of the input file. "
                f"Expected a dictionary in the messages list. Found {type(message)}",
                line_number=idx + 1,
                error_source="key_value",
            )
        for column in REQUIRED_COLUMNS_MESSAGE:
            if column not in message:
                raise InvalidFileFormatError(
                    message=f"Field `{column}` is missing for a turn `{message}` on line {idx + 1} "
                    "of the the input file.",
                    line_number=idx + 1,
                    error_source="key_value",
                )
            else:
                if not isinstance(message[column], str):
                    raise InvalidFileFormatError(
                        message=f"Invalid format on line {idx + 1} in the column {column} for turn `{message}` "
                        f"of the input file. Expected string. Found {type(message[column])}",
                        line_number=idx + 1,
                        error_source="text_field",
                    )

        if has_weights and "weight" in message:
            weight = message["weight"]
            if not isinstance(weight, int):
                raise InvalidFileFormatError(
                    message="Weight must be an integer",
                    line_number=idx + 1,
                    error_source="key_value",
                )
            if weight not in {0, 1}:
                raise InvalidFileFormatError(
                    message="Weight must be either 0 or 1",
                    line_number=idx + 1,
                    error_source="key_value",
                )
        if message["role"] not in POSSIBLE_ROLES_CONVERSATION:
            raise InvalidFileFormatError(
                message=f"Found invalid role `{message['role']}` in the messages on the line {idx + 1}. "
                f"Possible roles in the conversation are: {POSSIBLE_ROLES_CONVERSATION}",
                line_number=idx + 1,
                error_source="key_value",
            )

        if previous_role == message["role"]:
            raise InvalidFileFormatError(
                message=f"Invalid role turns on line {idx + 1} of the input file. "
                "`user` and `assistant` roles must alternate user/assistant/user/assistant/...",
                line_number=idx + 1,
                error_source="key_value",
            )
        previous_role = message["role"]


def validate_preference_openai(example: Dict[str, Any], idx: int = 0) -> None:
    """Validate the OpenAI preference dataset format.

    Args:
        example (dict): Input entry to be checked.
        idx (int): Line number in the file.

    Raises:
        InvalidFileFormatError: If the dataset format is invalid.
    """
    if not isinstance(example["input"], dict):
        raise InvalidFileFormatError(
            message="The dataset is malformed, the `input` field must be a dictionary.",
            line_number=idx + 1,
            error_source="key_value",
        )

    if "messages" not in example["input"]:
        raise InvalidFileFormatError(
            message="The dataset is malformed, the `input` dictionary must contain a `messages` field.",
            line_number=idx + 1,
            error_source="key_value",
        )

    validate_messages(example["input"]["messages"], idx)

    for output_field in ["preferred_output", "non_preferred_output"]:
        if not isinstance(example[output_field], list):
            raise InvalidFileFormatError(
                message=f"The dataset is malformed, the `{output_field}` field must be a list.",
                line_number=idx + 1,
                error_source="key_value",
            )

        if len(example[output_field]) != 1:
            raise InvalidFileFormatError(
                message=f"The dataset is malformed, the `{output_field}` list must contain exactly one message.",
                line_number=idx + 1,
                error_source="key_value",
            )
        if "role" not in example[output_field][0]:
            raise InvalidFileFormatError(
                message=f"The dataset is malformed, the `{output_field}` message is missing the `role` field.",
                line_number=idx + 1,
                error_source="key_value",
            )
        elif example[output_field][0]["role"] != "assistant":
            raise InvalidFileFormatError(
                message=f"The dataset is malformed, the `{output_field}` must contain an assistant message.",
                line_number=idx + 1,
                error_source="key_value",
            )

    validate_messages(example["preferred_output"], idx)
    validate_messages(example["non_preferred_output"], idx)


def _check_jsonl(file: Path) -> Dict[str, Any]:
    report_dict: Dict[str, Any] = {}
    # Check that the file is UTF-8 encoded. If not report where the error occurs.
    try:
        with file.open(encoding="utf-8") as f:
            f.read()
        report_dict["utf8"] = True
    except UnicodeDecodeError as e:
        report_dict["utf8"] = False
        report_dict["message"] = f"File is not UTF-8 encoded. Error raised: {e}."
        report_dict["is_check_passed"] = False
        return report_dict

    dataset_format = None
    with file.open() as f:
        idx = -1
        try:
            for idx, line in enumerate(f):
                json_line = json.loads(line)

                if not isinstance(json_line, dict):
                    raise InvalidFileFormatError(
                        message=(
                            f"Error parsing file. Invalid format on line {idx + 1} of the input file. "
                            "Datasets must follow text, conversational, or instruction format. For more"
                            "information, see https://docs.together.ai/docs/fine-tuning-data-preparation"
                        ),
                        line_number=idx + 1,
                        error_source="line_type",
                    )

                current_format = None
                for possible_format in JSONL_REQUIRED_COLUMNS_MAP:
                    if all(
                        column in json_line
                        for column in JSONL_REQUIRED_COLUMNS_MAP[possible_format]
                    ):
                        if current_format is None:
                            current_format = possible_format
                        elif current_format != possible_format:
                            raise InvalidFileFormatError(
                                message="Found multiple dataset formats in the input file. "
                                f"Got {current_format} and {possible_format} on line {idx + 1}.",
                                line_number=idx + 1,
                                error_source="format",
                            )

                        # Check that there are no extra columns
                        for column in json_line:
                            if (
                                column
                                not in JSONL_REQUIRED_COLUMNS_MAP[possible_format]
                            ):
                                raise InvalidFileFormatError(
                                    message=f'Found extra column "{column}" in the line {idx + 1}.',
                                    line_number=idx + 1,
                                    error_source="format",
                                )

                if current_format is None:
                    raise InvalidFileFormatError(
                        message=(
                            f"Error parsing file. Could not detect a format for the line {idx + 1} with the columns:\n"
                            f"{json_line.keys()}"
                        ),
                        line_number=idx + 1,
                        error_source="format",
                    )
                if current_format == DatasetFormat.PREFERENCE_OPENAI:
                    validate_preference_openai(json_line, idx)
                elif current_format == DatasetFormat.CONVERSATION:
                    message_column = JSONL_REQUIRED_COLUMNS_MAP[
                        DatasetFormat.CONVERSATION
                    ][0]
                    validate_messages(json_line[message_column], idx)
                else:
                    for column in JSONL_REQUIRED_COLUMNS_MAP[current_format]:
                        if not isinstance(json_line[column], str):
                            raise InvalidFileFormatError(
                                message=f'Invalid value type for "{column}" key on line {idx + 1}. '
                                f"Expected string. Found {type(json_line[column])}.",
                                line_number=idx + 1,
                                error_source="key_value",
                            )

                if dataset_format is None:
                    dataset_format = current_format
                elif current_format is not None:
                    if current_format != dataset_format:
                        raise InvalidFileFormatError(
                            message="All samples in the dataset must have the same dataset format. "
                            f"Got {dataset_format} for the first line and {current_format} "
                            f"for the line {idx + 1}.",
                            line_number=idx + 1,
                            error_source="format",
                        )

            if idx + 1 < MIN_SAMPLES:
                report_dict["has_min_samples"] = False
                report_dict["message"] = (
                    f"Processing {file} resulted in only {idx + 1} samples. "
                    f"Our minimum is {MIN_SAMPLES} samples. "
                )
                report_dict["is_check_passed"] = False
            else:
                report_dict["num_samples"] = idx + 1
                report_dict["has_min_samples"] = True
                report_dict["is_check_passed"] = True

            report_dict["load_json"] = True

        except InvalidFileFormatError as e:
            report_dict["load_json"] = False
            report_dict["is_check_passed"] = False
            report_dict["message"] = e.message
            if e.line_number is not None:
                report_dict["line_number"] = e.line_number
            if e.error_source is not None:
                report_dict[e.error_source] = False
        except ValueError:
            report_dict["load_json"] = False
            if idx < 0:
                report_dict["message"] = (
                    "Unable to decode file. "
                    "File may be empty or in an unsupported format. "
                )
            else:
                report_dict["message"] = (
                    f"Error parsing json payload. Unexpected format on line {idx + 1}."
                )
            report_dict["is_check_passed"] = False

    if "text_field" not in report_dict:
        report_dict["text_field"] = True
    if "line_type" not in report_dict:
        report_dict["line_type"] = True
    if "key_value" not in report_dict:
        report_dict["key_value"] = True
    return report_dict


def _check_parquet(file: Path) -> Dict[str, Any]:
    try:
        # Pyarrow is optional as it's large (~80MB) and isn't compatible with older systems.
        from pyarrow import ArrowInvalid, parquet
    except ImportError:
        raise ImportError(
            "pyarrow is not installed and is required to use parquet files. Please install it via `pip install together[pyarrow]`"
        )

    report_dict: Dict[str, Any] = {}

    try:
        table = parquet.read_table(str(file), memory_map=True)
    except ArrowInvalid:
        report_dict["load_parquet"] = (
            f"An exception has occurred when loading the Parquet file {file}. Please check the file for corruption. "
            f"Exception trace:\n{format_exc()}"
        )
        report_dict["is_check_passed"] = False
        return report_dict

    column_names = table.schema.names
    if "input_ids" not in column_names:
        report_dict["load_parquet"] = (
            f"Parquet file {file} does not contain the `input_ids` column."
        )
        report_dict["is_check_passed"] = False
        return report_dict

    for column_name in column_names:
        if column_name not in PARQUET_EXPECTED_COLUMNS:
            report_dict["load_parquet"] = (
                f"Parquet file {file} contains an unexpected column {column_name}. "
                f"Only columns {PARQUET_EXPECTED_COLUMNS} are supported."
            )
            report_dict["is_check_passed"] = False
            return report_dict

    num_samples = len(table)
    if num_samples < MIN_SAMPLES:
        report_dict["has_min_samples"] = False
        report_dict["message"] = (
            f"Processing {file} resulted in only {num_samples} samples. "
            f"Our minimum is {MIN_SAMPLES} samples. "
        )
        report_dict["is_check_passed"] = False
        return report_dict
    else:
        report_dict["num_samples"] = num_samples

    report_dict["is_check_passed"] = True

    return report_dict



from functools import cached_property

from together.resources.audio.speech import AsyncSpeech, Speech
from together.types import (
    TogetherClient,
)


class Audio:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    @cached_property
    def speech(self) -> Speech:
        return Speech(self._client)


class AsyncAudio:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    @cached_property
    def speech(self) -> AsyncSpeech:
        return AsyncSpeech(self._client)



from __future__ import annotations

import json
import sys
from functools import wraps
from typing import Any, Callable, Dict, List, Literal, TypeVar, Union

import click

from together import Together
from together.error import InvalidRequestError
from together.types import DedicatedEndpoint, ListEndpoint


def print_endpoint(
    endpoint: Union[DedicatedEndpoint, ListEndpoint],
) -> None:
    """Print endpoint details in a Docker-like format or JSON."""

    # Print header info
    click.echo(f"ID:\t\t{endpoint.id}")
    click.echo(f"Name:\t\t{endpoint.name}")

    # Print type-specific fields
    if isinstance(endpoint, DedicatedEndpoint):
        click.echo(f"Display Name:\t{endpoint.display_name}")
        click.echo(f"Hardware:\t{endpoint.hardware}")
        click.echo(
            f"Autoscaling:\tMin={endpoint.autoscaling.min_replicas}, "
            f"Max={endpoint.autoscaling.max_replicas}"
        )

    click.echo(f"Model:\t\t{endpoint.model}")
    click.echo(f"Type:\t\t{endpoint.type}")
    click.echo(f"Owner:\t\t{endpoint.owner}")
    click.echo(f"State:\t\t{endpoint.state}")
    click.echo(f"Created:\t{endpoint.created_at}")


F = TypeVar("F", bound=Callable[..., Any])


def print_api_error(
    e: InvalidRequestError,
) -> None:
    error_details = e.api_response.message

    if error_details and (
        "credentials" in error_details.lower()
        or "authentication" in error_details.lower()
    ):
        click.echo("Error: Invalid API key or authentication failed", err=True)
    else:
        click.echo(f"Error: {error_details}", err=True)


def handle_api_errors(f: F) -> F:
    """Decorator to handle common API errors in CLI commands."""

    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except InvalidRequestError as e:
            print_api_error(e)
            sys.exit(1)
        except Exception as e:
            click.echo(f"Error: An unexpected error occurred - {str(e)}", err=True)
            sys.exit(1)

    return wrapper  # type: ignore


@click.group()
@click.pass_context
def endpoints(ctx: click.Context) -> None:
    """Endpoints API commands"""
    pass


@endpoints.command()
@click.option(
    "--model",
    required=True,
    help="The model to deploy (e.g. meta-llama/Llama-4-Scout-17B-16E-Instruct)",
)
@click.option(
    "--min-replicas",
    type=int,
    default=1,
    help="Minimum number of replicas to deploy",
)
@click.option(
    "--max-replicas",
    type=int,
    default=1,
    help="Maximum number of replicas to deploy",
)
@click.option(
    "--gpu",
    type=click.Choice(["h100", "a100", "l40", "l40s", "rtx-6000"]),
    required=True,
    help="GPU type to use for inference",
)
@click.option(
    "--gpu-count",
    type=int,
    default=1,
    help="Number of GPUs to use per replica",
)
@click.option(
    "--display-name",
    help="A human-readable name for the endpoint",
)
@click.option(
    "--no-prompt-cache",
    is_flag=True,
    help="Disable the prompt cache for this endpoint",
)
@click.option(
    "--no-speculative-decoding",
    is_flag=True,
    help="Disable speculative decoding for this endpoint",
)
@click.option(
    "--no-auto-start",
    is_flag=True,
    help="Create the endpoint in STOPPED state instead of auto-starting it",
)
@click.option(
    "--inactive-timeout",
    type=int,
    help="Number of minutes of inactivity after which the endpoint will be automatically stopped. Set to 0 to disable.",
)
@click.option(
    "--wait",
    is_flag=True,
    default=True,
    help="Wait for the endpoint to be ready after creation",
)
@click.pass_obj
@handle_api_errors
def create(
    client: Together,
    model: str,
    min_replicas: int,
    max_replicas: int,
    gpu: str,
    gpu_count: int,
    display_name: str | None,
    no_prompt_cache: bool,
    no_speculative_decoding: bool,
    no_auto_start: bool,
    inactive_timeout: int | None,
    wait: bool,
) -> None:
    """Create a new dedicated inference endpoint."""
    # Map GPU types to their full hardware ID names
    gpu_map = {
        "h100": "nvidia_h100_80gb_sxm",
        "a100": "nvidia_a100_80gb_pcie" if gpu_count == 1 else "nvidia_a100_80gb_sxm",
        "l40": "nvidia_l40",
        "l40s": "nvidia_l40s",
        "rtx-6000": "nvidia_rtx_6000_ada",
    }

    hardware_id = f"{gpu_count}x_{gpu_map[gpu]}"

    try:
        response = client.endpoints.create(
            model=model,
            hardware=hardware_id,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
            display_name=display_name,
            disable_prompt_cache=no_prompt_cache,
            disable_speculative_decoding=no_speculative_decoding,
            state="STOPPED" if no_auto_start else "STARTED",
            inactive_timeout=inactive_timeout,
        )
    except InvalidRequestError as e:
        print_api_error(e)
        if "check the hardware api" in str(e).lower():
            fetch_and_print_hardware_options(
                client=client, model=model, print_json=False, available=True
            )

        sys.exit(1)

    # Print detailed information to stderr
    click.echo("Created dedicated endpoint with:", err=True)
    click.echo(f"  Model: {model}", err=True)
    click.echo(f"  Min replicas: {min_replicas}", err=True)
    click.echo(f"  Max replicas: {max_replicas}", err=True)
    click.echo(f"  Hardware: {hardware_id}", err=True)
    if display_name:
        click.echo(f"  Display name: {display_name}", err=True)
    if no_prompt_cache:
        click.echo("  Prompt cache: disabled", err=True)
    if no_speculative_decoding:
        click.echo("  Speculative decoding: disabled", err=True)
    if no_auto_start:
        click.echo("  Auto-start: disabled", err=True)
    if inactive_timeout is not None:
        click.echo(f"  Inactive timeout: {inactive_timeout} minutes", err=True)

    click.echo(f"Endpoint created successfully, id: {response.id}", err=True)

    if wait:
        import time

        click.echo("Waiting for endpoint to be ready...", err=True)
        while client.endpoints.get(response.id).state != "STARTED":
            time.sleep(1)
        click.echo("Endpoint ready", err=True)

    # Print only the endpoint ID to stdout
    click.echo(response.id)


@endpoints.command()
@click.argument("endpoint-id", required=True)
@click.option("--json", is_flag=True, help="Print output in JSON format")
@click.pass_obj
@handle_api_errors
def get(client: Together, endpoint_id: str, json: bool) -> None:
    """Get a dedicated inference endpoint."""
    endpoint = client.endpoints.get(endpoint_id)
    if json:
        import json as json_lib

        click.echo(json_lib.dumps(endpoint.model_dump(), indent=2))
    else:
        print_endpoint(endpoint)


@endpoints.command()
@click.option("--model", help="Filter hardware options by model")
@click.option("--json", is_flag=True, help="Print output in JSON format")
@click.option(
    "--available",
    is_flag=True,
    help="Print only available hardware options (can only be used if model is passed in)",
)
@click.pass_obj
@handle_api_errors
def hardware(client: Together, model: str | None, json: bool, available: bool) -> None:
    """List all available hardware options, optionally filtered by model."""
    fetch_and_print_hardware_options(client, model, json, available)


def fetch_and_print_hardware_options(
    client: Together, model: str | None, print_json: bool, available: bool
) -> None:
    """Print hardware options for a model."""

    message = "Available hardware options:" if available else "All hardware options:"
    click.echo(message, err=True)
    hardware_options = client.endpoints.list_hardware(model)
    if available:
        hardware_options = [
            hardware
            for hardware in hardware_options
            if hardware.availability is not None
            and hardware.availability.status == "available"
        ]

    if print_json:
        json_output = [hardware.model_dump() for hardware in hardware_options]
        click.echo(json.dumps(json_output, indent=2))
    else:
        for hardware in hardware_options:
            click.echo(f"  {hardware.id}", err=True)


@endpoints.command()
@click.argument("endpoint-id", required=True)
@click.option(
    "--wait", is_flag=True, default=True, help="Wait for the endpoint to stop"
)
@click.pass_obj
@handle_api_errors
def stop(client: Together, endpoint_id: str, wait: bool) -> None:
    """Stop a dedicated inference endpoint."""
    client.endpoints.update(endpoint_id, state="STOPPED")
    click.echo("Successfully marked endpoint as stopping", err=True)

    if wait:
        import time

        click.echo("Waiting for endpoint to stop...", err=True)
        while client.endpoints.get(endpoint_id).state != "STOPPED":
            time.sleep(1)
        click.echo("Endpoint stopped", err=True)

    click.echo(endpoint_id)


@endpoints.command()
@click.argument("endpoint-id", required=True)
@click.option(
    "--wait", is_flag=True, default=True, help="Wait for the endpoint to start"
)
@click.pass_obj
@handle_api_errors
def start(client: Together, endpoint_id: str, wait: bool) -> None:
    """Start a dedicated inference endpoint."""
    client.endpoints.update(endpoint_id, state="STARTED")
    click.echo("Successfully marked endpoint as starting", err=True)

    if wait:
        import time

        click.echo("Waiting for endpoint to start...", err=True)
        while client.endpoints.get(endpoint_id).state != "STARTED":
            time.sleep(1)
        click.echo("Endpoint started", err=True)

    click.echo(endpoint_id)


@endpoints.command()
@click.argument("endpoint-id", required=True)
@click.pass_obj
@handle_api_errors
def delete(client: Together, endpoint_id: str) -> None:
    """Delete a dedicated inference endpoint."""
    client.endpoints.delete(endpoint_id)
    click.echo("Successfully deleted endpoint", err=True)
    click.echo(endpoint_id)


@endpoints.command()
@click.option("--json", is_flag=True, help="Print output in JSON format")
@click.option(
    "--type",
    type=click.Choice(["dedicated", "serverless"]),
    help="Filter by endpoint type",
)
@click.pass_obj
@handle_api_errors
def list(
    client: Together, json: bool, type: Literal["dedicated", "serverless"] | None
) -> None:
    """List all inference endpoints (includes both dedicated and serverless endpoints)."""
    endpoints: List[ListEndpoint] = client.endpoints.list(type=type)

    if not endpoints:
        click.echo("No dedicated endpoints found", err=True)
        return

    click.echo("Endpoints:", err=True)
    if json:
        import json as json_lib

        click.echo(
            json_lib.dumps([endpoint.model_dump() for endpoint in endpoints], indent=2)
        )
    else:
        for endpoint in endpoints:
            print_endpoint(
                endpoint,
            )
            click.echo()


@endpoints.command()
@click.argument("endpoint-id", required=True)
@click.option(
    "--display-name",
    help="A new human-readable name for the endpoint",
)
@click.option(
    "--min-replicas",
    type=int,
    help="New minimum number of replicas to maintain",
)
@click.option(
    "--max-replicas",
    type=int,
    help="New maximum number of replicas to scale up to",
)
@click.option(
    "--inactive-timeout",
    type=int,
    help="Number of minutes of inactivity after which the endpoint will be automatically stopped. Set to 0 to disable.",
)
@click.pass_obj
@handle_api_errors
def update(
    client: Together,
    endpoint_id: str,
    display_name: str | None,
    min_replicas: int | None,
    max_replicas: int | None,
    inactive_timeout: int | None,
) -> None:
    """Update a dedicated inference endpoint's configuration."""
    if not any([display_name, min_replicas, max_replicas, inactive_timeout]):
        click.echo("Error: At least one update option must be specified", err=True)
        sys.exit(1)

    # If only one of min/max replicas is specified, we need both for the update
    if (min_replicas is None) != (max_replicas is None):
        click.echo(
            "Error: Both --min-replicas and --max-replicas must be specified together",
            err=True,
        )
        sys.exit(1)

    # Build kwargs for the update
    kwargs: Dict[str, Any] = {}
    if display_name is not None:
        kwargs["display_name"] = display_name
    if min_replicas is not None and max_replicas is not None:
        kwargs["min_replicas"] = min_replicas
        kwargs["max_replicas"] = max_replicas
    if inactive_timeout is not None:
        kwargs["inactive_timeout"] = inactive_timeout

    _response = client.endpoints.update(endpoint_id, **kwargs)

    # Print what was updated
    click.echo("Updated endpoint configuration:", err=True)
    if display_name:
        click.echo(f"  Display name: {display_name}", err=True)
    if min_replicas is not None and max_replicas is not None:
        click.echo(f"  Min replicas: {min_replicas}", err=True)
        click.echo(f"  Max replicas: {max_replicas}", err=True)
    if inactive_timeout is not None:
        click.echo(f"  Inactive timeout: {inactive_timeout} minutes", err=True)

    click.echo("Successfully updated endpoint", err=True)
    click.echo(endpoint_id)



from __future__ import annotations

from typing import List

from together.abstract import api_requestor
from together.together_response import TogetherResponse
from together.types import (
    ModelObject,
    TogetherClient,
    TogetherRequest,
)


class ModelsBase:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    def _filter_dedicated_models(
        self, models: List[ModelObject], dedicated_response: TogetherResponse
    ) -> List[ModelObject]:
        """
        Filter models based on dedicated model response.

        Args:
            models (List[ModelObject]): List of all models
            dedicated_response (TogetherResponse): Response from autoscale models endpoint

        Returns:
            List[ModelObject]: Filtered list of models
        """
        assert isinstance(dedicated_response.data, list)

        # Create a set of dedicated model names for efficient lookup
        dedicated_model_names = {model["name"] for model in dedicated_response.data}

        # Filter models to only include those in dedicated_model_names
        # Note: The model.id from ModelObject matches the name field in the autoscale response
        return [model for model in models if model.id in dedicated_model_names]


class Models(ModelsBase):
    def list(
        self,
        dedicated: bool = False,
    ) -> List[ModelObject]:
        """
        Method to return list of models on the API

        Args:
            dedicated (bool, optional): If True, returns only dedicated models. Defaults to False.

        Returns:
            List[ModelObject]: List of model objects
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url="models",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        assert isinstance(response.data, list)

        models = [ModelObject(**model) for model in response.data]

        if dedicated:
            # Get dedicated models
            dedicated_response, _, _ = requestor.request(
                options=TogetherRequest(
                    method="GET",
                    url="autoscale/models",
                ),
                stream=False,
            )

            models = self._filter_dedicated_models(models, dedicated_response)

        models.sort(key=lambda x: x.id.lower())

        return models


class AsyncModels(ModelsBase):
    async def list(
        self,
        dedicated: bool = False,
    ) -> List[ModelObject]:
        """
        Async method to return list of models on API

        Args:
            dedicated (bool, optional): If True, returns only dedicated models. Defaults to False.

        Returns:
            List[ModelObject]: List of model objects
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url="models",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        assert isinstance(response.data, list)

        models = [ModelObject(**model) for model in response.data]

        if dedicated:
            # Get dedicated models
            dedicated_response, _, _ = await requestor.arequest(
                options=TogetherRequest(
                    method="GET",
                    url="autoscale/models",
                ),
                stream=False,
            )

            models = self._filter_dedicated_models(models, dedicated_response)

        models.sort(key=lambda x: x.id.lower())

        return models



from __future__ import annotations

from enum import Enum
from typing import Optional
from datetime import datetime

from pydantic import Field

from together.types.abstract import BaseModel


class BatchJobStatus(str, Enum):
    """
    The status of a batch job
    """

    VALIDATING = "VALIDATING"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    EXPIRED = "EXPIRED"
    CANCELLED = "CANCELLED"


class BatchEndpoint(str, Enum):
    """
    The endpoint of a batch job
    """

    COMPLETIONS = "/v1/completions"
    CHAT_COMPLETIONS = "/v1/chat/completions"
    # More endpoints can be added here as needed


class BatchJob(BaseModel):
    """
    A batch job object
    """

    id: str
    user_id: str
    input_file_id: str
    file_size_bytes: int
    status: BatchJobStatus
    job_deadline: datetime
    created_at: datetime
    endpoint: str
    progress: float = 0.0
    model_id: Optional[str] = None
    output_file_id: Optional[str] = None
    error_file_id: Optional[str] = None
    error: Optional[str] = None
    completed_at: Optional[datetime] = None



from __future__ import annotations

import json
import os
import sys
import platform
from typing import TYPE_CHECKING, Any, Dict


if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem

import together
from together import error
from together.utils._log import _console_log_level
from together.utils import log_info


def get_headers(
    method: str | None = None,
    api_key: str | None = None,
    extra: "SupportsKeysAndGetItem[str, Any] | None" = None,
) -> Dict[str, str]:
    """
    Generates request headers with API key, metadata, and supplied headers

    Args:
        method (str, optional): HTTP request type (POST, GET, etc.)
            Defaults to None.
        api_key (str, optional): API key to add as an Authorization header.
            Defaults to None.
        extra (SupportsKeysAndGetItem[str, Any], optional): Additional headers to add to request.
            Defaults to None.

    Returns:
        headers (Dict[str, str]): Compiled headers from data
    """

    user_agent = "Together/v1 PythonBindings/%s" % (together.version,)

    uname_without_node = " ".join(
        v for k, v in platform.uname()._asdict().items() if k != "node"
    )
    ua = {
        "bindings_version": together.version,
        "httplib": "requests",
        "lang": "python",
        "lang_version": platform.python_version(),
        "platform": platform.platform(),
        "publisher": "together",
        "uname": uname_without_node,
    }

    headers: Dict[str, Any] = {
        "X-Together-Client-User-Agent": json.dumps(ua),
        "Authorization": f"Bearer {default_api_key(api_key)}",
        "User-Agent": user_agent,
    }

    if _console_log_level():
        headers["Together-Debug"] = _console_log_level()
    if extra:
        headers.update(extra)

    return headers


def default_api_key(api_key: str | None = None) -> str | None:
    """
    API key fallback logic from input argument and environment variable

    Args:
        api_key (str, optional): Supplied API key. This argument takes priority over env var

    Returns:
        together_api_key (str): Returns API key from supplied input or env var

    Raises:
        together.error.AuthenticationError: if API key not found
    """
    if api_key:
        return api_key
    if os.environ.get("TOGETHER_API_KEY"):
        return os.environ.get("TOGETHER_API_KEY")

    raise error.AuthenticationError(together.constants.MISSING_API_KEY_MESSAGE)


def get_google_colab_secret(secret_name: str = "TOGETHER_API_KEY") -> str | None:
    """
    Checks to see if the user is running in Google Colab, and looks for the Together API Key secret.

    Args:
        secret_name (str, optional). Defaults to TOGETHER_API_KEY

    Returns:
        str: if the API key is found; None if an error occurred or the secret was not found.
    """
    # If running in Google Colab, check for Together in notebook secrets
    if "google.colab" in sys.modules:
        if TYPE_CHECKING:
            from google.colab import userdata  # type: ignore
        else:
            from google.colab import userdata

        try:
            api_key = userdata.get(secret_name)
            if not isinstance(api_key, str):
                return None
            else:
                return str(api_key)
        except userdata.NotebookAccessError:
            log_info(
                "The TOGETHER_API_KEY Colab secret was found, but notebook access is disabled. Please enable notebook "
                "access for the secret."
            )
        except userdata.SecretNotFoundError:
            # warn and carry on
            log_info("Colab: No Google Colab secret named TOGETHER_API_KEY was found.")

        return None

    else:
        return None



from __future__ import annotations

from typing import Dict, List, Literal, Optional, Union

from together.abstract import api_requestor
from together.together_response import TogetherResponse
from together.types import TogetherClient, TogetherRequest
from together.types.endpoints import DedicatedEndpoint, HardwareWithStatus, ListEndpoint


class Endpoints:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    def list(
        self, type: Optional[Literal["dedicated", "serverless"]] = None
    ) -> List[ListEndpoint]:
        """
        List all endpoints, can be filtered by type.

        Args:
            type (str, optional): Filter endpoints by type ("dedicated" or "serverless"). Defaults to None.

        Returns:
            List[ListEndpoint]: List of endpoint objects
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        params = {}
        if type is not None:
            params["type"] = type

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url="endpoints",
                params=params,
            ),
            stream=False,
        )

        response.data = response.data["data"]

        assert isinstance(response, TogetherResponse)
        assert isinstance(response.data, list)

        return [ListEndpoint(**endpoint) for endpoint in response.data]

    def create(
        self,
        *,
        model: str,
        hardware: str,
        min_replicas: int,
        max_replicas: int,
        display_name: Optional[str] = None,
        disable_prompt_cache: bool = False,
        disable_speculative_decoding: bool = False,
        state: Literal["STARTED", "STOPPED"] = "STARTED",
        inactive_timeout: Optional[int] = None,
    ) -> DedicatedEndpoint:
        """
        Create a new dedicated endpoint.

        Args:
            model (str): The model to deploy on this endpoint
            hardware (str): The hardware configuration to use for this endpoint
            min_replicas (int): The minimum number of replicas to maintain
            max_replicas (int): The maximum number of replicas to scale up to
            display_name (str, optional): A human-readable name for the endpoint
            disable_prompt_cache (bool, optional): Whether to disable the prompt cache. Defaults to False.
            disable_speculative_decoding (bool, optional): Whether to disable speculative decoding. Defaults to False.
            state (str, optional): The desired state of the endpoint. Defaults to "STARTED".
            inactive_timeout (int, optional): The number of minutes of inactivity after which the endpoint will be automatically stopped. Set to 0 to disable automatic timeout.

        Returns:
            DedicatedEndpoint: Object containing endpoint information
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        data: Dict[str, Union[str, bool, Dict[str, int], int]] = {
            "model": model,
            "hardware": hardware,
            "autoscaling": {
                "min_replicas": min_replicas,
                "max_replicas": max_replicas,
            },
            "disable_prompt_cache": disable_prompt_cache,
            "disable_speculative_decoding": disable_speculative_decoding,
            "state": state,
        }

        if display_name is not None:
            data["display_name"] = display_name

        if inactive_timeout is not None:
            data["inactive_timeout"] = inactive_timeout

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="POST",
                url="endpoints",
                params=data,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return DedicatedEndpoint(**response.data)

    def get(self, endpoint_id: str) -> DedicatedEndpoint:
        """
        Get details of a specific endpoint.

        Args:
            endpoint_id (str): ID of the endpoint to retrieve

        Returns:
            DedicatedEndpoint: Object containing endpoint information
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url=f"endpoints/{endpoint_id}",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return DedicatedEndpoint(**response.data)

    def delete(self, endpoint_id: str) -> None:
        """
        Delete a specific endpoint.

        Args:
            endpoint_id (str): ID of the endpoint to delete
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        requestor.request(
            options=TogetherRequest(
                method="DELETE",
                url=f"endpoints/{endpoint_id}",
            ),
            stream=False,
        )

    def update(
        self,
        endpoint_id: str,
        *,
        min_replicas: Optional[int] = None,
        max_replicas: Optional[int] = None,
        state: Optional[Literal["STARTED", "STOPPED"]] = None,
        display_name: Optional[str] = None,
        inactive_timeout: Optional[int] = None,
    ) -> DedicatedEndpoint:
        """
        Update an endpoint's configuration.

        Args:
            endpoint_id (str): ID of the endpoint to update
            min_replicas (int, optional): The minimum number of replicas to maintain
            max_replicas (int, optional): The maximum number of replicas to scale up to
            state (str, optional): The desired state of the endpoint ("STARTED" or "STOPPED")
            display_name (str, optional): A human-readable name for the endpoint
            inactive_timeout (int, optional): The number of minutes of inactivity after which the endpoint will be automatically stopped. Set to 0 to disable automatic timeout.

        Returns:
            DedicatedEndpoint: Object containing endpoint information
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        data: Dict[str, Union[str, Dict[str, int], int]] = {}

        if min_replicas is not None or max_replicas is not None:
            current_min = min_replicas
            current_max = max_replicas
            if current_min is None or current_max is None:
                # Get current values if only one is specified
                current = self.get(endpoint_id=endpoint_id)
                current_min = current_min or current.autoscaling.min_replicas
                current_max = current_max or current.autoscaling.max_replicas
            data["autoscaling"] = {
                "min_replicas": current_min,
                "max_replicas": current_max,
            }

        if state is not None:
            data["state"] = state

        if display_name is not None:
            data["display_name"] = display_name

        if inactive_timeout is not None:
            data["inactive_timeout"] = inactive_timeout

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="PATCH",
                url=f"endpoints/{endpoint_id}",
                params=data,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return DedicatedEndpoint(**response.data)

    def list_hardware(self, model: Optional[str] = None) -> List[HardwareWithStatus]:
        """
        List available hardware configurations.

        Args:
            model (str, optional): Filter hardware configurations by model compatibility. When provided,
                                 the response includes availability status for each compatible configuration.

        Returns:
            List[HardwareWithStatus]: List of hardware configurations with their status
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        params = {}
        if model is not None:
            params["model"] = model

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url="hardware",
                params=params,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        assert isinstance(response.data, dict)
        assert isinstance(response.data["data"], list)

        return [HardwareWithStatus(**item) for item in response.data["data"]]


class AsyncEndpoints:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    async def list(
        self, type: Optional[Literal["dedicated", "serverless"]] = None
    ) -> List[ListEndpoint]:
        """
        List all endpoints, can be filtered by type.

        Args:
            type (str, optional): Filter endpoints by type ("dedicated" or "serverless"). Defaults to None.

        Returns:
            List[ListEndpoint]: List of endpoint objects
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        params = {}
        if type is not None:
            params["type"] = type

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url="endpoints",
                params=params,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        assert isinstance(response.data, list)

        return [ListEndpoint(**endpoint) for endpoint in response.data]

    async def create(
        self,
        *,
        model: str,
        hardware: str,
        min_replicas: int,
        max_replicas: int,
        display_name: Optional[str] = None,
        disable_prompt_cache: bool = False,
        disable_speculative_decoding: bool = False,
        state: Literal["STARTED", "STOPPED"] = "STARTED",
        inactive_timeout: Optional[int] = None,
    ) -> DedicatedEndpoint:
        """
        Create a new dedicated endpoint.

        Args:
            model (str): The model to deploy on this endpoint
            hardware (str): The hardware configuration to use for this endpoint
            min_replicas (int): The minimum number of replicas to maintain
            max_replicas (int): The maximum number of replicas to scale up to
            display_name (str, optional): A human-readable name for the endpoint
            disable_prompt_cache (bool, optional): Whether to disable the prompt cache. Defaults to False.
            disable_speculative_decoding (bool, optional): Whether to disable speculative decoding. Defaults to False.
            state (str, optional): The desired state of the endpoint. Defaults to "STARTED".
            inactive_timeout (int, optional): The number of minutes of inactivity after which the endpoint will be automatically stopped. Set to 0 to disable automatic timeout.

        Returns:
            DedicatedEndpoint: Object containing endpoint information
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        data: Dict[str, Union[str, bool, Dict[str, int], int]] = {
            "model": model,
            "hardware": hardware,
            "autoscaling": {
                "min_replicas": min_replicas,
                "max_replicas": max_replicas,
            },
            "disable_prompt_cache": disable_prompt_cache,
            "disable_speculative_decoding": disable_speculative_decoding,
            "state": state,
        }

        if display_name is not None:
            data["display_name"] = display_name

        if inactive_timeout is not None:
            data["inactive_timeout"] = inactive_timeout

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="POST",
                url="endpoints",
                params=data,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return DedicatedEndpoint(**response.data)

    async def get(self, endpoint_id: str) -> DedicatedEndpoint:
        """
        Get details of a specific endpoint.

        Args:
            endpoint_id (str): ID of the endpoint to retrieve

        Returns:
            DedicatedEndpoint: Object containing endpoint information
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url=f"endpoints/{endpoint_id}",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return DedicatedEndpoint(**response.data)

    async def delete(self, endpoint_id: str) -> None:
        """
        Delete a specific endpoint.

        Args:
            endpoint_id (str): ID of the endpoint to delete
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        await requestor.arequest(
            options=TogetherRequest(
                method="DELETE",
                url=f"endpoints/{endpoint_id}",
            ),
            stream=False,
        )

    async def update(
        self,
        endpoint_id: str,
        *,
        min_replicas: Optional[int] = None,
        max_replicas: Optional[int] = None,
        state: Optional[Literal["STARTED", "STOPPED"]] = None,
        display_name: Optional[str] = None,
        inactive_timeout: Optional[int] = None,
    ) -> DedicatedEndpoint:
        """
        Update an endpoint's configuration.

        Args:
            endpoint_id (str): ID of the endpoint to update
            min_replicas (int, optional): The minimum number of replicas to maintain
            max_replicas (int, optional): The maximum number of replicas to scale up to
            state (str, optional): The desired state of the endpoint ("STARTED" or "STOPPED")
            display_name (str, optional): A human-readable name for the endpoint
            inactive_timeout (int, optional): The number of minutes of inactivity after which the endpoint will be automatically stopped. Set to 0 to disable automatic timeout.

        Returns:
            DedicatedEndpoint: Object containing endpoint information
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        data: Dict[str, Union[str, Dict[str, int], int]] = {}

        if min_replicas is not None or max_replicas is not None:
            current_min = min_replicas
            current_max = max_replicas
            if current_min is None or current_max is None:
                # Get current values if only one is specified
                current = await self.get(endpoint_id=endpoint_id)
                current_min = current_min or current.autoscaling.min_replicas
                current_max = current_max or current.autoscaling.max_replicas
            data["autoscaling"] = {
                "min_replicas": current_min,
                "max_replicas": current_max,
            }

        if state is not None:
            data["state"] = state

        if display_name is not None:
            data["display_name"] = display_name

        if inactive_timeout is not None:
            data["inactive_timeout"] = inactive_timeout

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="PATCH",
                url=f"endpoints/{endpoint_id}",
                params=data,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        return DedicatedEndpoint(**response.data)

    async def list_hardware(
        self, model: Optional[str] = None
    ) -> List[HardwareWithStatus]:
        """
        List available hardware configurations.

        Args:
            model (str, optional): Filter hardware configurations by model compatibility. When provided,
                                 the response includes availability status for each compatible configuration.

        Returns:
            List[HardwareWithStatus]: List of hardware configurations with their status
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        params = {}
        if model is not None:
            params["model"] = model

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url="hardware",
                params=params,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        assert isinstance(response.data, dict)
        assert isinstance(response.data["data"], list)

        return [HardwareWithStatus(**item) for item in response.data["data"]]



from together.types.abstract import TogetherClient
from together.types.audio_speech import (
    AudioLanguage,
    AudioResponseEncoding,
    AudioResponseFormat,
    AudioSpeechRequest,
    AudioSpeechStreamChunk,
    AudioSpeechStreamEvent,
    AudioSpeechStreamResponse,
)
from together.types.chat_completions import (
    ChatCompletionChunk,
    ChatCompletionRequest,
    ChatCompletionResponse,
)
from together.types.common import TogetherRequest
from together.types.completions import (
    CompletionChunk,
    CompletionRequest,
    CompletionResponse,
)
from together.types.embeddings import EmbeddingRequest, EmbeddingResponse
from together.types.endpoints import Autoscaling, DedicatedEndpoint, ListEndpoint
from together.types.files import (
    FileDeleteResponse,
    FileList,
    FileObject,
    FilePurpose,
    FileRequest,
    FileResponse,
    FileType,
)
from together.types.finetune import (
    TrainingMethodDPO,
    TrainingMethodSFT,
    FinetuneCheckpoint,
    CosineLRScheduler,
    CosineLRSchedulerArgs,
    FinetuneDownloadResult,
    LinearLRScheduler,
    LinearLRSchedulerArgs,
    FinetuneLRScheduler,
    FinetuneList,
    FinetuneListEvents,
    FinetuneRequest,
    FinetuneResponse,
    FinetuneTrainingLimits,
    FullTrainingType,
    LoRATrainingType,
    TrainingType,
)
from together.types.images import ImageRequest, ImageResponse
from together.types.models import ModelObject
from together.types.rerank import RerankRequest, RerankResponse
from together.types.batch import BatchJob, BatchJobStatus, BatchEndpoint


__all__ = [
    "TogetherClient",
    "TogetherRequest",
    "CompletionChunk",
    "CompletionRequest",
    "CompletionResponse",
    "ChatCompletionChunk",
    "ChatCompletionRequest",
    "ChatCompletionResponse",
    "EmbeddingRequest",
    "EmbeddingResponse",
    "FinetuneCheckpoint",
    "FinetuneRequest",
    "FinetuneResponse",
    "FinetuneList",
    "FinetuneListEvents",
    "FinetuneDownloadResult",
    "FinetuneLRScheduler",
    "LinearLRScheduler",
    "LinearLRSchedulerArgs",
    "CosineLRScheduler",
    "CosineLRSchedulerArgs",
    "FileRequest",
    "FileResponse",
    "FileList",
    "FileDeleteResponse",
    "FileObject",
    "FilePurpose",
    "FileType",
    "ImageRequest",
    "ImageResponse",
    "ModelObject",
    "TrainingType",
    "FullTrainingType",
    "LoRATrainingType",
    "TrainingMethodDPO",
    "TrainingMethodSFT",
    "RerankRequest",
    "RerankResponse",
    "FinetuneTrainingLimits",
    "AudioSpeechRequest",
    "AudioResponseFormat",
    "AudioLanguage",
    "AudioResponseEncoding",
    "AudioSpeechStreamChunk",
    "AudioSpeechStreamEvent",
    "AudioSpeechStreamResponse",
    "DedicatedEndpoint",
    "ListEndpoint",
    "Autoscaling",
    "BatchJob",
    "BatchJobStatus",
    "BatchEndpoint",
]



from __future__ import annotations

from typing import Any, AsyncGenerator, Dict, Iterator, List, Union

from together.abstract import api_requestor
from together.together_response import TogetherResponse
from together.types import (
    AudioSpeechRequest,
    AudioResponseFormat,
    AudioLanguage,
    AudioResponseEncoding,
    AudioSpeechStreamChunk,
    AudioSpeechStreamEvent,
    AudioSpeechStreamResponse,
    TogetherClient,
    TogetherRequest,
)


class Speech:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    def create(
        self,
        *,
        model: str,
        input: str,
        voice: str | None = None,
        response_format: str = "wav",
        language: str = "en",
        response_encoding: str = "pcm_f32le",
        sample_rate: int = 44100,
        stream: bool = False,
        **kwargs: Any,
    ) -> AudioSpeechStreamResponse:
        """
        Method to generate audio from input text using a specified model.

        Args:
            model (str): The name of the model to query.
            input (str): Input text to generate the audio for.
            voice (str, optional): The voice to use for generating the audio.
                Defaults to None.
            response_format (str, optional): The format of audio output.
                Defaults to "wav".
            language (str, optional): Language of input text.
                Defaults to "en".
            response_encoding (str, optional): Audio encoding of response.
                Defaults to "pcm_f32le".
            sample_rate (int, optional): Sampling rate to use for the output audio.
                Defaults to 44100.
            stream (bool, optional): If true, output is streamed for several characters at a time.
                Defaults to False.

        Returns:
            Union[bytes, Iterator[AudioSpeechStreamChunk]]: The generated audio as bytes or an iterator over audio stream chunks.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        parameter_payload = AudioSpeechRequest(
            model=model,
            input=input,
            voice=voice,
            response_format=AudioResponseFormat(response_format),
            language=AudioLanguage(language),
            response_encoding=AudioResponseEncoding(response_encoding),
            sample_rate=sample_rate,
            stream=stream,
            **kwargs,
        ).model_dump(exclude_none=True)

        response, streamed, _ = requestor.request(
            options=TogetherRequest(
                method="POST",
                url="audio/speech",
                params=parameter_payload,
            ),
            stream=stream,
        )

        return AudioSpeechStreamResponse(response=response)


class AsyncSpeech:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    async def create(
        self,
        *,
        model: str,
        input: str,
        voice: str | None = None,
        response_format: str = "wav",
        language: str = "en",
        response_encoding: str = "pcm_f32le",
        sample_rate: int = 44100,
        stream: bool = False,
        **kwargs: Any,
    ) -> AudioSpeechStreamResponse:
        """
        Async method to generate audio from input text using a specified model.

        Args:
            model (str): The name of the model to query.
            input (str): Input text to generate the audio for.
            voice (str, optional): The voice to use for generating the audio.
                Defaults to None.
            response_format (str, optional): The format of audio output.
                Defaults to "wav".
            language (str, optional): Language of input text.
                Defaults to "en".
            response_encoding (str, optional): Audio encoding of response.
                Defaults to "pcm_f32le".
            sample_rate (int, optional): Sampling rate to use for the output audio.
                Defaults to 44100.
            stream (bool, optional): If true, output is streamed for several characters at a time.
                Defaults to False.

        Returns:
            Union[bytes, AsyncGenerator[AudioSpeechStreamChunk, None]]: The generated audio as bytes or an async generator over audio stream chunks.
        """

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        parameter_payload = AudioSpeechRequest(
            model=model,
            input=input,
            voice=voice,
            response_format=AudioResponseFormat(response_format),
            language=AudioLanguage(language),
            response_encoding=AudioResponseEncoding(response_encoding),
            sample_rate=sample_rate,
            stream=stream,
            **kwargs,
        ).model_dump(exclude_none=True)

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="POST",
                url="audio/speech",
                params=parameter_payload,
            ),
            stream=stream,
        )

        return AudioSpeechStreamResponse(response=response)



from __future__ import annotations

from enum import Enum
from typing import Iterator
import threading

from pydantic import BaseModel, ConfigDict

from together.together_response import TogetherResponse
import base64


class AudioResponseFormat(str, Enum):
    MP3 = "mp3"
    WAV = "wav"
    RAW = "raw"


class AudioLanguage(str, Enum):
    EN = "en"
    DE = "de"
    FR = "fr"
    ES = "es"
    HI = "hi"
    IT = "it"
    JA = "ja"
    KO = "ko"
    NL = "nl"
    PL = "pl"
    PT = "pt"
    RU = "ru"
    SV = "sv"
    TR = "tr"
    ZH = "zh"


class AudioResponseEncoding(str, Enum):
    PCM_F32LE = "pcm_f32le"
    PCM_S16LE = "pcm_s16le"
    PCM_MULAW = "pcm_mulaw"
    PCM_ALAW = "pcm_alaw"


class AudioObjectType(str, Enum):
    AUDIO_TTS_CHUNK = "audio.tts.chunk"


class StreamSentinelType(str, Enum):
    DONE = "[DONE]"


class AudioSpeechRequest(BaseModel):
    model: str
    input: str
    voice: str | None = None
    response_format: AudioResponseFormat = AudioResponseFormat.MP3
    language: AudioLanguage = AudioLanguage.EN
    response_encoding: AudioResponseEncoding = AudioResponseEncoding.PCM_F32LE
    sample_rate: int = 44100
    stream: bool = False


class AudioSpeechStreamChunk(BaseModel):
    object: AudioObjectType = AudioObjectType.AUDIO_TTS_CHUNK
    model: str
    b64: str


class AudioSpeechStreamEvent(BaseModel):
    data: AudioSpeechStreamChunk


class StreamSentinel(BaseModel):
    data: StreamSentinelType = StreamSentinelType.DONE


class AudioSpeechStreamEventResponse(BaseModel):
    response: AudioSpeechStreamEvent | StreamSentinel


class AudioSpeechStreamResponse(BaseModel):

    response: TogetherResponse | Iterator[TogetherResponse]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def stream_to_file(self, file_path: str) -> None:

        if isinstance(self.response, TogetherResponse):
            # save response to file
            with open(file_path, "wb") as f:
                f.write(self.response.data)

        elif isinstance(self.response, Iterator):

            with open(file_path, "wb") as f:
                for chunk in self.response:

                    # Try to parse as stream chunk
                    stream_event_response = AudioSpeechStreamEventResponse(
                        response={"data": chunk.data}
                    )

                    if isinstance(stream_event_response.response, StreamSentinel):
                        break

                    # decode base64
                    audio = base64.b64decode(stream_event_response.response.data.b64)

                    f.write(audio)



from __future__ import annotations

from typing import Any, Dict, Literal, Union

from pydantic import Field

from together.types.endpoints import TogetherJSONModel


class FileInput(TogetherJSONModel):
    """File input to be uploaded to the code interpreter session."""

    name: str = Field(description="The name of the file.")
    encoding: Literal["string", "base64"] = Field(
        description="Encoding of the file content. Use 'string' for text files and 'base64' for binary files."
    )
    content: str = Field(description="The content of the file, encoded as specified.")


class InterpreterOutput(TogetherJSONModel):
    """Base class for interpreter output types."""

    type: Literal["stdout", "stderr", "error", "display_data", "execute_result"] = (
        Field(description="The type of output")
    )
    data: Union[str, Dict[str, Any]] = Field(description="The output data")


class ExecuteResponseData(TogetherJSONModel):
    """Data from code execution response."""

    outputs: list[InterpreterOutput] = Field(
        description="List of outputs from execution", default_factory=list
    )
    errors: Union[str, None] = Field(
        description="Any errors that occurred during execution", default=None
    )
    session_id: str = Field(
        description="Identifier of the current session. Used to make follow-up calls."
    )
    status: str = Field(description="Status of the execution", default="completed")


class ExecuteResponse(TogetherJSONModel):
    """Response from code execution."""

    data: ExecuteResponseData = Field(
        description="The response data containing outputs and session information"
    )


__all__ = [
    "FileInput",
    "InterpreterOutput",
    "ExecuteResponseData",
    "ExecuteResponse",
]



from __future__ import annotations

import json
import re
from datetime import datetime, timezone
from textwrap import wrap
from typing import Any, Literal

import click
from click.core import ParameterSource  # type: ignore[attr-defined]
from rich import print as rprint
from tabulate import tabulate

from together import Together
from together.cli.api.utils import BOOL_WITH_AUTO, INT_WITH_MAX
from together.types.finetune import (
    DownloadCheckpointType,
    FinetuneEventType,
    FinetuneTrainingLimits,
)
from together.utils import (
    finetune_price_to_dollars,
    format_timestamp,
    log_warn,
    log_warn_once,
    parse_timestamp,
)


_CONFIRMATION_MESSAGE = (
    "You are about to create a fine-tuning job. "
    "The cost of your job will be determined by the model size, the number of tokens "
    "in the training file, the number of tokens in the validation file, the number of epochs, and "
    "the number of evaluations. Visit https://www.together.ai/pricing to get a price estimate.\n"
    "You can pass `-y` or `--confirm` to your command to skip this message.\n\n"
    "Do you want to proceed?"
)


class DownloadCheckpointTypeChoice(click.Choice):
    def __init__(self) -> None:
        super().__init__([ct.value for ct in DownloadCheckpointType])

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.Context | None
    ) -> DownloadCheckpointType:
        value = super().convert(value, param, ctx)
        return DownloadCheckpointType(value)


@click.group(name="fine-tuning")
@click.pass_context
def fine_tuning(ctx: click.Context) -> None:
    """Fine-tunes API commands"""
    pass


@fine_tuning.command()
@click.pass_context
@click.option(
    "--training-file",
    "-t",
    type=str,
    required=True,
    help="Training file ID from Files API",
)
@click.option("--model", "-m", type=str, help="Base model name")
@click.option(
    "--n-epochs", "-ne", type=int, default=1, help="Number of epochs to train for"
)
@click.option(
    "--validation-file", type=str, default="", help="Validation file ID from Files API"
)
@click.option("--n-evals", type=int, default=0, help="Number of evaluation loops")
@click.option(
    "--n-checkpoints", "-c", type=int, default=1, help="Number of checkpoints to save"
)
@click.option(
    "--batch-size", "-b", type=INT_WITH_MAX, default="max", help="Train batch size"
)
@click.option("--learning-rate", "-lr", type=float, default=1e-5, help="Learning rate")
@click.option(
    "--lr-scheduler-type",
    type=click.Choice(["linear", "cosine"]),
    default="cosine",
    help="Learning rate scheduler type",
)
@click.option(
    "--min-lr-ratio",
    type=float,
    default=0.0,
    help="The ratio of the final learning rate to the peak learning rate",
)
@click.option(
    "--scheduler-num-cycles",
    type=float,
    default=0.5,
    help="Number or fraction of cycles for the cosine learning rate scheduler.",
)
@click.option(
    "--warmup-ratio",
    type=float,
    default=0.0,
    help="Warmup ratio for the learning rate scheduler.",
)
@click.option(
    "--max-grad-norm",
    type=float,
    default=1.0,
    help="Max gradient norm to be used for gradient clipping. Set to 0 to disable.",
)
@click.option(
    "--weight-decay",
    type=float,
    default=0.0,
    help="Weight decay",
)
@click.option(
    "--lora/--no-lora",
    type=bool,
    default=True,
    help="Whether to use LoRA adapters for fine-tuning",
)
@click.option("--lora-r", type=int, default=8, help="LoRA adapters' rank")
@click.option("--lora-dropout", type=float, default=0, help="LoRA adapters' dropout")
@click.option("--lora-alpha", type=float, default=8, help="LoRA adapters' alpha")
@click.option(
    "--lora-trainable-modules",
    type=str,
    default="all-linear",
    help="Trainable modules for LoRA adapters. For example, 'all-linear', 'q_proj,v_proj'",
)
@click.option(
    "--training-method",
    type=click.Choice(["sft", "dpo"]),
    default="sft",
    help="Training method to use. Options: sft (supervised fine-tuning), dpo (Direct Preference Optimization)",
)
@click.option(
    "--dpo-beta",
    type=float,
    default=None,
    help="Beta parameter for DPO training (only used when '--training-method' is 'dpo')",
)
@click.option(
    "--dpo-normalize-logratios-by-length",
    type=bool,
    default=False,
    help=(
        "Whether to normalize logratios by sample length "
        "(only used when '--training-method' is 'dpo')"
    ),
)
@click.option(
    "--rpo-alpha",
    type=float,
    default=None,
    help=(
        "RPO alpha parameter of DPO training to include NLL in the loss "
        "(only used when '--training-method' is 'dpo')"
    ),
)
@click.option(
    "--simpo-gamma",
    type=float,
    default=None,
    help="SimPO gamma parameter (only used when '--training-method' is 'dpo')",
)
@click.option(
    "--suffix",
    "-s",
    type=str,
    default=None,
    help="Suffix for the fine-tuned model name",
)
@click.option("--wandb-api-key", type=str, default=None, help="Wandb API key")
@click.option("--wandb-base-url", type=str, default=None, help="Wandb base URL")
@click.option("--wandb-project-name", type=str, default=None, help="Wandb project name")
@click.option("--wandb-name", type=str, default=None, help="Wandb run name")
@click.option(
    "--confirm",
    "-y",
    type=bool,
    is_flag=True,
    default=False,
    help="Whether to skip the launch confirmation message",
)
@click.option(
    "--train-on-inputs",
    type=BOOL_WITH_AUTO,
    default=None,
    help="Whether to mask the user messages in conversational data or prompts in instruction data. "
    "`auto` will automatically determine whether to mask the inputs based on the data format.",
)
@click.option(
    "--from-checkpoint",
    type=str,
    default=None,
    help="The checkpoint identifier to continue training from a previous fine-tuning job. "
    "The format: {$JOB_ID/$OUTPUT_MODEL_NAME}:{$STEP}. "
    "The step value is optional, without it the final checkpoint will be used.",
)
@click.option(
    "--hf-api-token",
    type=str,
    default=None,
    help="HF API token to use for uploading a checkpoint to a private repo",
)
@click.option(
    "--hf-output-repo-name",
    type=str,
    default=None,
    help="HF repo to upload the fine-tuned model to",
)
def create(
    ctx: click.Context,
    training_file: str,
    validation_file: str,
    model: str,
    n_epochs: int,
    n_evals: int,
    n_checkpoints: int,
    batch_size: int | Literal["max"],
    learning_rate: float,
    lr_scheduler_type: Literal["linear", "cosine"],
    min_lr_ratio: float,
    scheduler_num_cycles: float,
    warmup_ratio: float,
    max_grad_norm: float,
    weight_decay: float,
    lora: bool,
    lora_r: int,
    lora_dropout: float,
    lora_alpha: float,
    lora_trainable_modules: str,
    suffix: str,
    wandb_api_key: str,
    wandb_base_url: str,
    wandb_project_name: str,
    wandb_name: str,
    confirm: bool,
    train_on_inputs: bool | Literal["auto"],
    training_method: str,
    dpo_beta: float | None,
    dpo_normalize_logratios_by_length: bool,
    rpo_alpha: float | None,
    simpo_gamma: float | None,
    from_checkpoint: str,
    hf_api_token: str | None,
    hf_output_repo_name: str | None,
) -> None:
    """Start fine-tuning"""
    client: Together = ctx.obj

    training_args: dict[str, Any] = dict(
        training_file=training_file,
        model=model,
        n_epochs=n_epochs,
        validation_file=validation_file,
        n_evals=n_evals,
        n_checkpoints=n_checkpoints,
        batch_size=batch_size,
        learning_rate=learning_rate,
        lr_scheduler_type=lr_scheduler_type,
        min_lr_ratio=min_lr_ratio,
        scheduler_num_cycles=scheduler_num_cycles,
        warmup_ratio=warmup_ratio,
        max_grad_norm=max_grad_norm,
        weight_decay=weight_decay,
        lora=lora,
        lora_r=lora_r,
        lora_dropout=lora_dropout,
        lora_alpha=lora_alpha,
        lora_trainable_modules=lora_trainable_modules,
        suffix=suffix,
        wandb_api_key=wandb_api_key,
        wandb_base_url=wandb_base_url,
        wandb_project_name=wandb_project_name,
        wandb_name=wandb_name,
        train_on_inputs=train_on_inputs,
        training_method=training_method,
        dpo_beta=dpo_beta,
        dpo_normalize_logratios_by_length=dpo_normalize_logratios_by_length,
        rpo_alpha=rpo_alpha,
        simpo_gamma=simpo_gamma,
        from_checkpoint=from_checkpoint,
        hf_api_token=hf_api_token,
        hf_output_repo_name=hf_output_repo_name,
    )

    if model is None and from_checkpoint is None:
        raise click.BadParameter("You must specify either a model or a checkpoint")

    model_name = model
    if from_checkpoint is not None:
        model_name = from_checkpoint.split(":")[0]

    model_limits: FinetuneTrainingLimits = client.fine_tuning.get_model_limits(
        model=model_name,
    )

    if lora:
        if model_limits.lora_training is None:
            raise click.BadParameter(
                f"LoRA fine-tuning is not supported for the model `{model}`"
            )
        if training_method == "dpo":
            default_batch_size = model_limits.lora_training.max_batch_size_dpo
        else:
            default_batch_size = model_limits.lora_training.max_batch_size
        default_values = {
            "lora_r": model_limits.lora_training.max_rank,
            "batch_size": default_batch_size,
            "learning_rate": 1e-3,
        }

        for arg in default_values:
            arg_source = ctx.get_parameter_source("arg")  # type: ignore[attr-defined]
            if arg_source == ParameterSource.DEFAULT:
                training_args[arg] = default_values[arg_source]

        if ctx.get_parameter_source("lora_alpha") == ParameterSource.DEFAULT:  # type: ignore[attr-defined]
            training_args["lora_alpha"] = training_args["lora_r"] * 2
    else:
        if model_limits.full_training is None:
            raise click.BadParameter(
                f"Full fine-tuning is not supported for the model `{model}`"
            )

        for param in ["lora_r", "lora_dropout", "lora_alpha", "lora_trainable_modules"]:
            param_source = ctx.get_parameter_source(param)  # type: ignore[attr-defined]
            if param_source != ParameterSource.DEFAULT:
                raise click.BadParameter(
                    f"You set LoRA parameter `{param}` for a full fine-tuning job. "
                    f"Please change the job type with --lora or remove `{param}` from the arguments"
                )

        batch_size_source = ctx.get_parameter_source("batch_size")  # type: ignore[attr-defined]
        if batch_size_source == ParameterSource.DEFAULT:
            if training_method == "dpo":
                training_args["batch_size"] = (
                    model_limits.full_training.max_batch_size_dpo
                )
            else:
                training_args["batch_size"] = model_limits.full_training.max_batch_size

    if n_evals <= 0 and validation_file:
        log_warn(
            "Warning: You have specified a validation file but the number of evaluation loops is set to 0. No evaluations will be performed."
        )
    elif n_evals > 0 and not validation_file:
        raise click.BadParameter(
            "You have specified a number of evaluation loops but no validation file."
        )

    if confirm or click.confirm(_CONFIRMATION_MESSAGE, default=True, show_default=True):
        response = client.fine_tuning.create(
            **training_args,
            verbose=True,
        )

        report_string = f"Successfully submitted a fine-tuning job {response.id}"
        if response.created_at is not None:
            created_time = datetime.strptime(
                response.created_at, "%Y-%m-%dT%H:%M:%S.%f%z"
            )
            # created_at reports UTC time, we use .astimezone() to convert to local time
            formatted_time = created_time.astimezone().strftime("%m/%d/%Y, %H:%M:%S")
            report_string += f" at {formatted_time}"
        rprint(report_string)
    else:
        click.echo("No confirmation received, stopping job launch")


@fine_tuning.command()
@click.pass_context
def list(ctx: click.Context) -> None:
    """List fine-tuning jobs"""
    client: Together = ctx.obj

    response = client.fine_tuning.list()

    response.data = response.data or []

    # Use a default datetime for None values to make sure the key function always returns a comparable value
    epoch_start = datetime.fromtimestamp(0, tz=timezone.utc)
    response.data.sort(key=lambda x: parse_timestamp(x.created_at or "") or epoch_start)

    display_list = []
    for i in response.data:
        display_list.append(
            {
                "Fine-tune ID": i.id,
                "Model Output Name": "\n".join(wrap(i.output_name or "", width=30)),
                "Status": i.status,
                "Created At": i.created_at,
                "Price": f"""${
                    finetune_price_to_dollars(float(str(i.total_price)))
                }""",  # convert to string for mypy typing
            }
        )
    table = tabulate(display_list, headers="keys", tablefmt="grid", showindex=True)

    click.echo(table)


@fine_tuning.command()
@click.pass_context
@click.argument("fine_tune_id", type=str, required=True)
def retrieve(ctx: click.Context, fine_tune_id: str) -> None:
    """Retrieve fine-tuning job details"""
    client: Together = ctx.obj

    response = client.fine_tuning.retrieve(fine_tune_id)

    # remove events from response for cleaner output
    response.events = None

    click.echo(json.dumps(response.model_dump(exclude_none=True), indent=4))


@fine_tuning.command()
@click.pass_context
@click.argument("fine_tune_id", type=str, required=True)
@click.option(
    "--quiet", is_flag=True, help="Do not prompt for confirmation before cancelling job"
)
def cancel(ctx: click.Context, fine_tune_id: str, quiet: bool = False) -> None:
    """Cancel fine-tuning job"""
    client: Together = ctx.obj
    if not quiet:
        confirm_response = input(
            "You will be billed for any completed training steps upon cancellation. "
            f"Do you want to cancel job {fine_tune_id}? [y/N]"
        )
        if "y" not in confirm_response.lower():
            click.echo({"status": "Cancel not submitted"})
            return
    response = client.fine_tuning.cancel(fine_tune_id)

    click.echo(json.dumps(response.model_dump(exclude_none=True), indent=4))


@fine_tuning.command()
@click.pass_context
@click.argument("fine_tune_id", type=str, required=True)
def list_events(ctx: click.Context, fine_tune_id: str) -> None:
    """List fine-tuning events"""
    client: Together = ctx.obj

    response = client.fine_tuning.list_events(fine_tune_id)

    response.data = response.data or []

    display_list = []
    for i in response.data:
        display_list.append(
            {
                "Message": "\n".join(wrap(i.message or "", width=50)),
                "Type": i.type,
                "Created At": parse_timestamp(i.created_at or ""),
                "Hash": i.hash,
            }
        )
    table = tabulate(display_list, headers="keys", tablefmt="grid", showindex=True)

    click.echo(table)


@fine_tuning.command()
@click.pass_context
@click.argument("fine_tune_id", type=str, required=True)
def list_checkpoints(ctx: click.Context, fine_tune_id: str) -> None:
    """List available checkpoints for a fine-tuning job"""
    client: Together = ctx.obj

    checkpoints = client.fine_tuning.list_checkpoints(fine_tune_id)

    display_list = []
    for checkpoint in checkpoints:
        display_list.append(
            {
                "Type": checkpoint.type,
                "Timestamp": format_timestamp(checkpoint.timestamp),
                "Name": checkpoint.name,
            }
        )

    if display_list:
        click.echo(f"Job {fine_tune_id} contains the following checkpoints:")
        table = tabulate(display_list, headers="keys", tablefmt="grid")
        click.echo(table)
        click.echo("\nTo download a checkpoint, use `together fine-tuning download`")
    else:
        click.echo(f"No checkpoints found for job {fine_tune_id}")


@fine_tuning.command()
@click.pass_context
@click.argument("fine_tune_id", type=str, required=True)
@click.option(
    "--output_dir",
    "-o",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    required=False,
    default=None,
    help="Output directory",
)
@click.option(
    "--checkpoint-step",
    "-s",
    type=int,
    required=False,
    default=None,
    help="Download fine-tuning checkpoint. Defaults to latest.",
)
@click.option(
    "--checkpoint-type",
    type=DownloadCheckpointTypeChoice(),
    required=False,
    default=DownloadCheckpointType.DEFAULT.value,
    help="Specifies checkpoint type. 'merged' and 'adapter' options work only for LoRA jobs.",
)
def download(
    ctx: click.Context,
    fine_tune_id: str,
    output_dir: str,
    checkpoint_step: int | None,
    checkpoint_type: DownloadCheckpointType,
) -> None:
    """Download fine-tuning checkpoint"""
    client: Together = ctx.obj

    response = client.fine_tuning.download(
        fine_tune_id,
        output=output_dir,
        checkpoint_step=checkpoint_step,
        checkpoint_type=checkpoint_type,
    )

    click.echo(json.dumps(response.model_dump(exclude_none=True), indent=4))



import enum

# Session constants
TIMEOUT_SECS = 600
MAX_SESSION_LIFETIME_SECS = 180
MAX_CONNECTION_RETRIES = 2
MAX_RETRIES = 5
INITIAL_RETRY_DELAY = 0.5
MAX_RETRY_DELAY = 8.0

# API defaults
BASE_URL = "https://api.together.xyz/v1"

# Download defaults
DOWNLOAD_BLOCK_SIZE = 10 * 1024 * 1024  # 10 MB
DISABLE_TQDM = False

# Messages
MISSING_API_KEY_MESSAGE = """TOGETHER_API_KEY not found.
Please set it as an environment variable or set it as together.api_key
Find your TOGETHER_API_KEY at https://api.together.xyz/settings/api-keys"""

# Minimum number of samples required for fine-tuning file
MIN_SAMPLES = 1

# the number of bytes in a gigabyte, used to convert bytes to GB for readable comparison
NUM_BYTES_IN_GB = 2**30

# maximum number of GB sized files we support finetuning for
MAX_FILE_SIZE_GB = 4.9

# expected columns for Parquet files
PARQUET_EXPECTED_COLUMNS = ["input_ids", "attention_mask", "labels"]


class DatasetFormat(enum.Enum):
    """Dataset format enum."""

    GENERAL = "general"
    CONVERSATION = "conversation"
    INSTRUCTION = "instruction"
    PREFERENCE_OPENAI = "preference_openai"


JSONL_REQUIRED_COLUMNS_MAP = {
    DatasetFormat.GENERAL: ["text"],
    DatasetFormat.CONVERSATION: ["messages"],
    DatasetFormat.INSTRUCTION: ["prompt", "completion"],
    DatasetFormat.PREFERENCE_OPENAI: [
        "input",
        "preferred_output",
        "non_preferred_output",
    ],
}
REQUIRED_COLUMNS_MESSAGE = ["role", "content"]
POSSIBLE_ROLES_CONVERSATION = ["system", "user", "assistant"]



from __future__ import annotations

from typing import List

from together.abstract import api_requestor
from together.together_response import TogetherResponse
from together.types import (
    TogetherClient,
    TogetherRequest,
    BatchJob,
)


class Batches:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    def create_batch(self, file_id: str, endpoint: str) -> BatchJob:

        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        parameter_payload = {
            "input_file_id": file_id,
            "endpoint": endpoint,
            "completion_window": "24h",
        }

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="POST",
                url=f"batches",
                params=parameter_payload,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        response_body = response.data.get("job", {})
        return BatchJob(**response_body)

    def get_batch(self, batch_job_id: str) -> BatchJob:
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url=f"batches/{batch_job_id}",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        return BatchJob(**response.data)

    def list_batches(self) -> List[BatchJob]:
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="GET",
                url="batches",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        jobs = response.data or []
        return [BatchJob(**job) for job in jobs]


class AsyncBatches:
    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    async def create_batch(self, file_id: str, endpoint: str) -> BatchJob:
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        parameter_payload = {
            "input_file_id": file_id,
            "endpoint": endpoint,
            "completion_window": "24h",
        }

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="POST",
                url=f"batches",
                params=parameter_payload,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        response_body = response.data.get("job", {})
        return BatchJob(**response_body)

    async def get_batch(self, batch_job_id: str) -> BatchJob:
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url=f"batches/{batch_job_id}",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        return BatchJob(**response.data)

    async def list_batches(self) -> List[BatchJob]:
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        response, _, _ = await requestor.arequest(
            options=TogetherRequest(
                method="GET",
                url="batches",
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)
        jobs = response.data or []
        return [BatchJob(**job) for job in jobs]



import json as json_lib

import click
from tabulate import tabulate

from together import Together
from together.types.models import ModelObject


@click.group()
@click.pass_context
def models(ctx: click.Context) -> None:
    """Models API commands"""
    pass


@models.command()
@click.option(
    "--type",
    type=click.Choice(["dedicated"]),
    help="Filter models by type (dedicated: models that can be deployed as dedicated endpoints)",
)
@click.option(
    "--json",
    is_flag=True,
    help="Output in JSON format",
)
@click.pass_context
def list(ctx: click.Context, type: str | None, json: bool) -> None:
    """List models"""
    client: Together = ctx.obj

    response = client.models.list(dedicated=(type == "dedicated"))

    display_list = []

    model: ModelObject
    for model in response:
        display_list.append(
            {
                "ID": model.id,
                "Name": model.display_name,
                "Organization": model.organization,
                "Type": model.type,
                "Context Length": model.context_length,
                "License": model.license,
                "Input per 1M token": model.pricing.input,
                "Output per 1M token": model.pricing.output,
            }
        )

    if json:
        click.echo(json_lib.dumps(display_list, indent=2))
    else:
        click.echo(tabulate(display_list, headers="keys", tablefmt="plain"))



from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional
from pydantic import ValidationError

from together.abstract import api_requestor
from together.together_response import TogetherResponse
from together.types import TogetherClient, TogetherRequest
from together.types.code_interpreter import ExecuteResponse, FileInput


class CodeInterpreter:
    """Code Interpreter resource for executing code snippets."""

    def __init__(self, client: TogetherClient) -> None:
        self._client = client

    def run(
        self,
        code: str,
        language: Literal["python"],
        session_id: Optional[str] = None,
        files: Optional[List[Dict[str, Any]]] = None,
    ) -> ExecuteResponse:
        """Execute a code snippet, optionally with files.

        Args:
            code (str): Code snippet to execute
            language (str): Programming language for the code to execute. Currently only supports Python.
            session_id (str, optional): Identifier of the current session. Used to make follow-up calls.
            files (List[Dict], optional): Files to upload to the session before executing the code.

        Returns:
            ExecuteResponse: Object containing execution results and outputs

        Raises:
            ValidationError: If any dictionary in the `files` list does not conform to the
                required structure or types.
        """
        requestor = api_requestor.APIRequestor(
            client=self._client,
        )

        data: Dict[str, Any] = {
            "code": code,
            "language": language,
        }

        if session_id is not None:
            data["session_id"] = session_id

        if files is not None:
            serialized_files = []
            try:
                for file_dict in files:
                    # Validate the dictionary by creating a FileInput instance
                    validated_file = FileInput(**file_dict)
                    # Serialize the validated model back to a dict for the API call
                    serialized_files.append(validated_file.model_dump())
            except ValidationError as e:
                raise ValueError(f"Invalid file input format: {e}") from e
            except TypeError as e:
                raise ValueError(
                    f"Invalid file input: Each item in 'files' must be a dictionary. Error: {e}"
                ) from e

            data["files"] = serialized_files

        # Use absolute URL to bypass the /v1 prefix
        response, _, _ = requestor.request(
            options=TogetherRequest(
                method="POST",
                url="/tci/execute",
                params=data,
            ),
            stream=False,
        )

        assert isinstance(response, TogetherResponse)

        # Return the response data directly since our types match the API structure
        return ExecuteResponse(**response.data)
